# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module for licensing and license serve checks."""

import os
import socket
import subprocess
import time

from ansys.mapdl.core import LOG
from ansys.mapdl.core.errors import LicenseServerConnectionError
from ansys.mapdl.core.misc import threaded_daemon

try:
    from ansys.tools.path import get_ansys_path, version_from_path

    _HAS_ATP = True

except ModuleNotFoundError:
    _HAS_ATP = False


LOCALHOST = "127.0.0.1"
LIC_PATH_ENVAR = "ANSYSLIC_DIR"
LIC_FILE_ENVAR = "ANSYSLMD_LICENSE_FILE"
APP_NAME = "FEAT_ANSYS"  # TODO: We need to make sure this is the type of feature we need to checkout.
LIC_TO_CHECK = ["mech_1"]

LICENSES = {
    "ansys": "Ansys Mechanical Enterprise",
    "meba": "Ansys Mechanical Enterprise Solver",
    "mech_2": "Ansys Mechanical Premium",
    "mech_1": "Ansys Mechanical Pro",
    "preppost": "Mechanical Enterprise PrepPost",
}
ALLOWABLE_LICENSES = list(LICENSES)

## Regarding license checking.
# The available licenses we can check against are (in order of
# complete/comprehensiveness):
# 1. ``Ansys`` Enterprise license (the most complete)
# 2. ``meba`` Mechanical Enterprise license
# 3. ``mech_2`` Premium license
# 4. ``mech_1`` Pro license` (the most limited)
# To keep very general, and since we need just to be able to solve, we are going to check against the lower license (``mech_1``).

# TODO: Implement a warning for insufficient license rights.


class LicenseChecker:
    """License checker class.

    Two methods are used and exposed with the :func:`LicenseChecker.start` method:

    * Check the ``licdebug`` log file for errors.
    * Check the available mechanical licenses using ``ansysli_util`` executable.

    Parameters
    ----------
    timeout : float, optional
        Timeout for the licensing log file check.

    """

    def __init__(self, timeout=30, verbose=None):
        self._license_file_msg = []
        self._license_file_success = None

        self._license_checkout_msg = []
        self._license_checkout_success = None
        self._timeout = timeout

        if verbose is not None:
            raise DeprecationWarning(
                "The argument 'verbose' has been deprecated, please use loggers from the logging module."
            )

        self._lic_file_thread = None
        self._checkout_thread = None

        self._stop = False
        self._is_connected = False

    @property
    def stop(self):
        return self._stop

    @stop.setter
    def stop(self, value):
        if bool(value):
            LOG.debug("Attempting to stop the license checker.")
        self._stop = bool(value)

    @property
    def is_connected(self):
        return self._is_connected

    @is_connected.setter
    def is_connected(self, value):
        if bool(value):
            LOG.debug("PyMAPDL has connected to a MAPDL session.")
        self._is_connected = bool(value)

    @threaded_daemon
    def check_license_file(self):
        try:
            self._check_license_file(self._timeout)
        except Exception as error:
            self._license_file_success = False
            self._license_file_msg.append(str(error))
        else:
            self._license_file_success = True

    @threaded_daemon
    def checkout_license(self, host=None):
        try:
            self._check_mech_license_available(host)
        except Exception as error:
            self._license_checkout_success = False
            self._license_checkout_msg.append(str(error))
        else:
            self._license_checkout_success = True

    def start(self, license_file=True, checkout_license=False):
        """Start monitoring the license file and attempt a license checkout.

        Parameters
        ----------
        license_file : bool, optional
            Start the license file thread.
        checkout_license : bool, optional
            Start the checkout license thread.  By default this is
            disabled.

        """
        if license_file and _HAS_ATP:
            self._lic_file_thread = self.check_license_file()
        if checkout_license:
            self._checkout_thread = self.checkout_license()

    def wait(self):
        """Wait until the license checks are complete or have timed out."""
        if self._lic_file_thread is not None:
            self._lic_file_thread.join()
        if self._checkout_thread is not None:
            self._checkout_thread.join()

    def check(self):
        """Report if the license checkout or license check was successful.

        It first checks the output from the license file and later the
        output from the checkout process.

        Returns
        -------
        bool
            ``True`` When license successfully checked out, ``False``
            when license check failed and nothing to report.  Checkout
            failure will raise
            :class:`ansys.mapdl.core.errors.LicenseServerConnectionError``.

        Raises
        ------
        LicenseServerConnectionError
            If there were any errors during the license checkout or
            license file check.

        """
        if self._license_file_success:
            return True
        elif self._license_file_success is False:
            raise LicenseServerConnectionError("\n".join(self._license_file_msg))

        if self._license_checkout_success:
            return True
        elif self._license_checkout_success is False:
            raise LicenseServerConnectionError("\n".join(self._license_checkout_msg))

        return False  # pragma: no cover

    def _check_license_file(self, timeout=30, notify_at_second=5):  # pragma: no cover
        """Check the output of the license client log for connection error.

        Expect type of errors with 'DENIED' in the header such as:
        ```
        2021/09/06 22:39:38    DENIED              ansys                           21.2 (2021.0512)             1/0/0/0                 1/1/1/1   10268:FEAT_ANSYS:gayuso@AAPDDqVK5WqNLrt.win.ansys.com:winx64   7368:192.168.18.10
                    Request name ansys does not exist in the licensing pool.
                    Cannot connect to license server system.
                    The license server manager (lmgrd) has not been started yet,
                    the wrong port@host or license file is being used, or the
                    port or hostname in the license file has been changed.
                    Feature:       ansys
                    Server name:   192.168.18.10
                    License path:  1055@AAPDDqVK5WqNLrt;
                    FlexNet Licensing error:-15,578.  System Error: 10049 "WinSock: Invalid address"
        ```
        On the other side, successful license checkout will print `CHECKOUT       $APPNAME` in the file, such as:
        ```
        2021/09/20 12:59:49    CHECKOUT            FEAT_ANSYS                      21.2 (2021.0512)             1/1/1/1                 1/1/1/1   11704:FEAT_ANSYS:gayuso@AAPDDqVK5WqNLve.win.ansys.com:winx64   6884:192.168.56.1
        ```
        for `FEAT_ANSYS`.

        Parameters
        ----------
        timeout : int, optional
            Time to keep checking the license log file for errors. Default to 10 (seconds).

        Raises
        ------
        LicenseServerConnectionError
            If there is an error message in the license log file.
        TimeoutError
            Exceeded ``timeout`` while waiting for the license log file.

        """
        licdebug_file = os.path.join(
            get_ansys_license_debug_file_path(),
            get_ansys_license_debug_file_name(),
        )
        file_iterator = get_ansys_license_debug_file_tail(licdebug_file)

        if self._check_license_file_iterator(
            file_iterator, licdebug_file, timeout, notify_at_second
        ):
            return True

    def _check_license_file_iterator(
        self, file_iterator, licdebug_file, timeout=30, notify_at_second=5
    ):
        """Loop over iterator"""
        max_time = time.time() + timeout
        notification_time = time.time() + notify_at_second
        notification_bool = True
        while time.time() < max_time:
            if self.stop:  # pragma: no cover
                LOG.debug("The license checker has received a stop signal.")
                raise Exception("The license checker has been stopped.")

            if self.is_connected:  # pragma: no cover
                LOG.debug("The MAPDL session got connected. Stopping license check.")
                return True

            if (
                time.time() > notification_time and notification_bool
            ):  # pragma: no cover
                msg = (
                    "PyMAPDL is taking longer than expected to connect to an MAPDL session.\n"
                    "Checking if there are any available licenses..."
                )
                LOG.debug(msg)
                print(msg)
                notification_bool = False

            msg = next(file_iterator)
            if msg:
                LOG.debug(f"Output from {licdebug_file}:\n{msg}")

            if "DENIED" in msg:
                # read to the end of the file
                time.sleep(0.05)  # permit flush
                messages = [msg]
                while True:
                    msg = next(file_iterator).strip()
                    if not msg:
                        break
                    messages.append(msg)  # pragma: no cover

                raise LicenseServerConnectionError("\n".join(messages))

            if "CHECKOUT" in msg:
                # successful license checkout
                return True

        raise TimeoutError(
            f"Exceeded timeout of {timeout} seconds while examining:\n{licdebug_file}"
        )

    def _checkout_license(self, lic, host=None, port=2325):
        """Check if a license is available using the Ansys license utility.

        It uses it own process.

        Parameters
        ----------
        lic : str
            License type.  For example, ``"meba".  one of the following:

            1. ``"Ansys"`` Enterprise license (the most complete)
            2. ``"meba"`` Mechanical Enterprise license
            3. ``"mech_2"`` Premium license
            4. ``"mech_1"`` Pro license` (the most limited)
        host : str, optional
            Host to attempt to checkout license from.  When set, this
            overrides any settings from the default license path.
        port : int, optional
            Port on the host to connect to.  Only used when ``host`` is set.

        """
        if lic.lower() not in ALLOWABLE_LICENSES:  # pragma: no cover
            raise ValueError(f"Invalid license '{lic}'")

        ansysli_util_path = get_ansys_license_utility_path()

        if not os.path.isfile(ansysli_util_path):  # pragma: no cover
            raise FileNotFoundError(
                "Ansys licensing path exists but ansysli_util not found at:\n"
                f"{ansysli_util_path}"
            )

        # allow the specification of ip and port
        env = os.environ.copy()
        if host is not None and port is not None:  # pragma: no cover
            env["ANSYSLI_SERVERS"] = f"{host}:{port}"
            env["ANS_FLEXLM_DISABLE_DEFLICPATH"] = "TRUE"

        tstart = time.time()
        process = subprocess.Popen(
            f'"{ansysli_util_path}" -checkout {lic}',
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            env=env,
            shell=True,
        )
        output = process.stdout.read().decode()

        t_elap = time.time() - tstart
        LOG.debug(f"License check complete in {t_elap:.2} seconds.\n")
        LOG.debug(output)

        return output

    def _check_mech_license_available(
        self, host=None, licenses=None
    ):  # pragma: no cover
        """Check if there mechanical license available by running 'ansysli_util'.

        This uses the default configuration available to MAPDL.

        .. warning::
          This method to check the license server status is not stable.  It
          is very likely it will show license not available when it actually
          is.

        Parameters
        ----------
        host : str, optional
            Override the default license server.  By default, use the
            values available in the Ansys license file.
        license : str, list, optional
            A list or single license to check for.  One or more of the
            following:

            1. ``"Ansys"`` Enterprise license (the most complete)
            2. ``"meba"`` Mechanical Enterprise license
            3. ``"mech_2"`` Premium license
            4. ``"mech_1"`` Pro license` (the most limited)

        Returns
        -------
        bool
            Returns True when the license exists.

        Raises
        ------
        LicenseServerConnectionError
            When errors messages found in the output of the license file.

        """

        if licenses is None:
            licenses = LIC_TO_CHECK
        elif isinstance(licenses, str):
            licenses = [licenses]

        msg1 = "No such feature exists"
        msg2 = "The server is down or is not responsive."
        for each_license in licenses:
            output = self._checkout_license(each_license, host)
            if msg1 in output or msg2 in output:
                raise LicenseServerConnectionError(output)

        return True


def get_ansys_license_debug_file_tail(licdebug_file, start_timeout=10, debug=False):
    """Get each of the licdebug file messages.

    This method keeps the ``licdebug`` file open checking for complete messages.
    It yields one message at a time when called.

    Parameters
    ----------
    licdebug_file : str
        Path to the ``licdebug`` file.wh
    start_timeout : float, optional
        Maximum timeout to wait until the file exists.

    Yields
    ------
    msg : str
        Message formatted as a single string.

    """
    # wait until file exists
    max_time = time.time() + start_timeout
    while not os.path.isfile(licdebug_file):  # pragma: no cover
        time.sleep(0.01)
        if time.time() > max_time:
            raise TimeoutError(
                f"Exceeded {start_timeout} seconds while waiting for {licdebug_file}"
                " to exist."
            )

    with open(licdebug_file) as fid:
        # Going to the end of the file.
        if not debug:  # pragma: no cover
            fid.seek(0, 2)
        while True:
            lines = "".join(fid.readlines())
            yield lines


def get_ansys_license_directory():  # pragma: no cover
    """Get the path to the Ansys license directory"""

    # it's possible the user has specified the license as an env var
    ansyslic_dir = None
    if LIC_FILE_ENVAR in os.environ:
        ansyslmd_var = os.environ[LIC_FILE_ENVAR]
        if not os.path.isfile(ansyslmd_var):
            # likely license info
            ansyslic_dir = None
    else:
        ansyslic_dir = os.getenv(LIC_PATH_ENVAR)

    # env var may not be specified, check in the usual location
    if ansyslic_dir is None:
        if os.name == "nt":  # pragma: no cover
            ansyslic_dir = os.path.join(
                os.environ["ProgramFiles"],
                "ANSYS Inc",
                "Shared Files",
                "Licensing",
            )
        else:
            ansyslic_dir = "/usr/ansys_inc/shared_files/licensing"

        if not os.path.isdir(ansyslic_dir):
            raise FileNotFoundError(
                f"Unable to locate ANSYS licencing path at {ansyslic_dir}\n"
                f"Specify the {LIC_PATH_ENVAR}"
            )

    return ansyslic_dir


def get_ansys_license_debug_file_name():  # pragma: no cover
    """Get license client log file name.

    This file change the name according to the ANSYS version and the type of license requested (``$appname``).
    * For ANSYS version 22.1 and above: ``licdebug.$hostname.$appname.$version.out``
    * For ANSYS version 21.2 and below: ``licdebug.$appname.$version.out``

    where:
    * ``$hostname`` is the name of the machine.
    * ``$appname`` is the name of the feature used by the license client. Eg. 'FEAT_ANSYS'
    * ``$version`` is the version of ANSYS. Eg 211 for version 21.1.

    Returns
    -------
    str
        licdebug log file complete name.

    """
    # Licdebug name convention:
    # - For version 22.1 and above: `licdebug.$hostname.$appname.$version.out`
    # - For version 21.2 and below: `licdebug.$appname.$version.out`

    name = "licdebug"
    hostname = socket.gethostname()
    appname = APP_NAME
    # This is the type of license my client requests (Windows 10, 2021R2)
    version = version_from_path("mapdl", get_ansys_path(allow_input=False))
    ending = "out"

    if version < 221:
        parts = (name, appname, version, ending)
    else:
        parts = (name, hostname, appname, version, ending)

    return ".".join([str(each_part) for each_part in parts])


def get_ansys_license_debug_file_path():  # pragma: no cover
    """Get license client log (``licdebug``) path.

    This path is obtained from the correspondent env variable (OS
    dependent) and appending ``.ansys``.

    Returns
    -------
    str
        path of the license client log file.
    """
    if os.name == "nt":
        folder = os.getenv("TEMP")
    elif os.name == "posix":
        folder = os.getenv("HOME")
    else:
        raise OSError(f"Unsupported OS {os.name}")

    return os.path.join(folder, ".ansys")


def get_ansys_license_utility_path():  # pragma: no cover
    """Return the ansys licencing utilities path."""
    ansyslic_dir = get_ansys_license_directory()
    if os.name == "nt":
        ansysli_util_path = os.path.join(ansyslic_dir, "winx64", "ansysli_util.exe")
    else:
        ansysli_util_path = os.path.join(ansyslic_dir, "linx64", "ansysli_util")

    return ansysli_util_path



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module to support MAPDL CAD geometry"""
from functools import wraps
from typing import TYPE_CHECKING, Any, Iterable, List, Optional, Sequence, Tuple, Union

import numpy as np
from numpy.typing import NDArray

from ansys.mapdl.core import _HAS_PYVISTA, Mapdl
from ansys.mapdl.core.errors import VersionError

if _HAS_PYVISTA:
    import pyvista as pv

if TYPE_CHECKING:  # pragma: no cover
    from pyiges import Iges

from ansys.mapdl.core.misc import requires_package, run_as_prep7, supress_logging
from ansys.mapdl.core.plotting.theme import MapdlTheme

VALID_SELECTION_TYPE = ["S", "R", "A", "U"]
VALID_SELECTION_ENTYTY = ["VOLU", "AREA", "LINE", "KP", "ELEM", "NODE"]

from ansys.mapdl.core.mapdl_core import (
    DEBUG_LEVELS,
    VALID_SELECTION_ENTITY_TP,
    VALID_SELECTION_TYPE_TP,
)

FLST_LOOKUP = {
    "NODE": 1,  # node numbers
    "ELEM": 2,  # element numbers
    "KP": 3,  # keypoint numbers
    "LINE": 4,  # line numbers
    "AREA": 5,  # area numbers
    "VOLU": 6,  # volume numbers
    "TRACE": 7,  # trace points
    "COORD": 8,  # coordinate locations
}

VALID_TYPE_MSG = """- 'S' : Select a new set (default)
- 'R' : Reselect a set from the current set.
- 'A' : Additionally select a set and extend the current set.
- 'U' : Unselect a set from the current set.
"""

VERSION_ERROR = """
In PyMAPDL 0.66.0 and later, the new geometry module does not allow calls on
``geometry.keypoints``, ``geometry.lines``, or ```geometry.areas```.

You can activate the old API like this:
>>> mapdl.legacy_geometry = True

For more information, see `Mesh and geometry <https://mapdl.docs.pyansys.com/version/stable/user_guide/mesh_geometry.html>`_.
"""


if _HAS_PYVISTA:

    class Multiblock(pv.MultiBlock):
        def __call__(self, *args, **kwargs):
            raise VersionError(VERSION_ERROR)

        @wraps(pv.MultiBlock.plot)
        def plot(self, *args, **kwargs):
            color = kwargs.pop("color", "white")
            return super().plot(*args, color=color, theme=MapdlTheme(), **kwargs)

    @requires_package("pyvista")
    def merge_polydata(items: Iterable["pv.PolyData"]) -> "pv.PolyData":
        """Merge list of polydata or unstructured grids"""

        # lazy import here for faster module loading
        from vtkmodules.vtkFiltersCore import vtkAppendPolyData

        afilter = vtkAppendPolyData()
        for item in items:
            afilter.AddInputData(item)
            afilter.Update()

        return pv.wrap(afilter.GetOutput())


class Geometry:
    """Pythonic representation of MAPDL CAD geometry

    Contains advanced methods to extend geometry building and
    selection within MAPDL.

    """

    def __init__(self, mapdl: Mapdl):
        """Geometry manager

        Class to help to manage geometry representations in an
        :class:`Mapdl instance <ansys.mapdl.core.Mapdl>` instance.

        Parameters
        ----------
        mapdl : ansys.mapdl.core.Mapdl
            Mapdl instance which this class references to.
        """
        from ansys.mapdl.core.mapdl import MapdlBase

        if not isinstance(mapdl, MapdlBase):
            raise TypeError("Must be initialized using a gRPC MAPDL class")

        self._mapdl = mapdl
        self._keypoints_cache = None
        self._lines_cache = None
        self._log = self._mapdl._log

    def _set_log_level(self, level: DEBUG_LEVELS) -> None:
        return self._mapdl.set_log_level(level)

    @requires_package("pyiges")
    def _load_iges(self) -> "Iges":
        """Loads the iges file from MAPDL as a pyiges class"""
        # Lazy import here for speed and stability
        # possible to exclude this import in the future
        try:
            from pyiges import Iges
        except ImportError:
            raise ImportError(
                "Please install pyiges to use this feature with:\n" "pip install pyiges"
            )
        return Iges(self._mapdl._generate_iges())

    def _reset_cache(self) -> None:
        self._keypoints_cache = None
        self._lines_cache = None

    def __setitem__(self, key: Any, value: Any):
        raise NotImplementedError("This method has not been implemented yet")

    @requires_package("pyvista")
    def __getitem__(self, name: str):
        name = name.lower()
        if "kp" in name:
            return self.keypoints[name]
        elif "line" in name:
            return self.lines[name]
        elif "area" in name:
            return self.areas[name]
        elif "volume" in name:
            return self.volumes[name]
        else:
            raise ValueError(
                f"The entity named '{name}' does not exist as a geometry entity."
            )

    @property
    @requires_package("pyiges")
    def _keypoints(self) -> Tuple[NDArray, NDArray]:
        """Returns keypoints cache"""
        if self._keypoints_cache is None:
            self._keypoints_cache = self._load_keypoints()
        return self._keypoints_cache

    @property
    @requires_package("pyvista")
    def keypoints(self) -> "pv.MultiBlock":
        """Obtain the keypoints geometry.

        Obtain the selected keypoints as a :class:`pyvista.MultiBlock` object.

        Returns
        -------
        pv.MultiBlock

        Examples
        --------

        >>> mapdl.geometry.keypoints
        MultiBlock (0x147a78220)
          N Blocks    26
          X Bounds    -0.016, 0.016
          Y Bounds    -0.008, 0.018
          Z Bounds    -0.003, 0.015

        You can plot the entities:

        >>> mapdl.geometry.keypoints.plot()

        You can access the individual keypoints using indexing:

        >>> keypoint0 = mapdl.geometry.keypoints[0]
        >>> keypoint0
        pyvista_ndarray([ 0.     ,  0.01778, -0.00318])

        You can use the entity name:

        >>> kp1 = mapdl.geometry.keypoints["kp 1"]
        >>> kp1
        pyvista_ndarray([ 0.     ,  0.01778, -0.00318])

        You can iterate the different elements:

        >>> points = mapdl.geometry.keypoints
        >>> for each_point in points:
                print(each_point.points[0])
        pyvista_ndarray([ 0.     ,  0.01778, -0.00318])
        ...

        Alternatively, to iterate you can use :meth:`Mapdl.geometry.get_keypoints`.

        >>> points = mapdl.geometry.get_keypoints(return_as_array=True)
        >>> for each_point in points:
                print(each_point)
        [ 0.     ,  0.01778, -0.00318]
        ...

        """
        mb = Multiblock(self.get_keypoints(return_as_list=True))
        # Setting names
        for ind, each_block in enumerate(mb):
            mapdl_index = int(each_block["entity_num"][0])
            mb.set_block_name(index=ind, name=f"kp {mapdl_index}")
        return mb

    @requires_package("pyvista")
    def get_keypoints(
        self,
        return_as_list: bool = False,
        return_as_array: bool = False,
        return_ids_in_array: bool = False,
    ) -> Union[NDArray[Any], "pv.PolyData", List["pv.PolyData"]]:
        """Obtain the keypoints geometry.

        Obtain the selected keypoints as a :class:`pyvista.PolyData` object or
        a list of :class:`pyvista.PolyData`.

        Parameters
        ----------
        return_as_list : bool, optional
            Whether to return the data as a list. The default is ``False``, in which case the data is returned as a PyVista :class:`PolyData<pyvista.PolyData>` object.
        return_as_array : bool, optional
            Whether to return the data as a :class:`numpy array <numpy.ndarray>`. The default  is ``False``, in which case the data is returned as a PyVista :class:`PolyData <pyvista.PolyData>` object.
        return_ids_in_array : bool, optional
            Whether to return the keypoint IDs in the output array. The default is
            ``False``, in which case the data is as a :class:`numpy array <numpy.ndarray>` object with only the coordinates. This parameter is only valid when ``return_as_array=True``.

        Returns
        -------
        Union[NDArray[Any], pv.PolyData, list[pv.PolyData]]

        Examples
        --------

        Return a single merged mesh.

        >>> kps = mapdl.geometry.get_keypoints()
        >>> kps
        PolyData (0x147d5b580)
          N Cells:    26
          N Points:   26
          N Strips:   0
          X Bounds:   -1.588e-02, 1.588e-02
          Y Bounds:   -7.620e-03, 1.778e-02
          Z Bounds:   -3.180e-03, 1.524e-02
          N Arrays:   1

        Return a list of keypoints as individual grids

        >>> keypoints = mapdl.geometry.get_keypoints(return_as_list=True)
        >>> keypoints
        [PolyData (0x147ca4be...rrays:   1, PolyData (0x147d5b8e...rrays:   1, PolyData (0x1491a42e...rrays:   1, PolyData (0x1491a440...rrays:   1, PolyData (0x1491a47c...rrays:   1, PolyData (0x1491a470...rrays:   1, PolyData (0x1491a4b2...rrays:   1, PolyData (0x1491a4e2...rrays:   1, PolyData (0x1491a49a...rrays:   1, PolyData (0x1491a4f4...rrays:   1, PolyData (0x1491a458...rrays:   1, PolyData (0x1491a4b8...rrays:   1, PolyData (0x1491a4dc...rrays:   1, PolyData (0x1491a506...rrays:   1, ...]

        Return the keypoints coordinates as a numpy array:

        >>> keypoints = mapdl.geometry.get_keypoints(return_as_array=True)
        array([[ 0.00000000e+00,  1.77800000e-02, -3.18000000e-03],
               [ 0.00000000e+00, -7.62000000e-03, -3.18000000e-03],
               [ 1.58750000e-02,  1.77800000e-02, -3.18000000e-03],
               [ 1.58750000e-02, -7.62000000e-03, -3.18000000e-03],
               [ 0.00000000e+00, -7.62000000e-03,  0.00000000e+00],
        ...

        When returning an array, you can also choose to output the keypoint
        ids as the first column:

        >>> keypoints = mapdl.geometry.get_keypoints(return_ids_in_array=True)
        array([[ 1.00000000e+00,  0.00000000e+00,  1.77800000e-02,
                -3.18000000e-03],
               [ 2.00000000e+00,  0.00000000e+00, -7.62000000e-03,
                -3.18000000e-03],
               [ 3.00000000e+00,  1.58750000e-02,  1.77800000e-02,
                -3.18000000e-03],
               [ 4.00000000e+00,  1.58750000e-02, -7.62000000e-03,
                -3.18000000e-03],
               [ 5.00000000e+00,  0.00000000e+00, -7.62000000e-03,
                 0.00000000e+00],
               [ 6.00000000e+00,  1.58750000e-02, -7.62000000e-03,
        ...

        """
        if return_as_array or return_ids_in_array:
            if return_ids_in_array:
                keyp = np.array(self._keypoints[0])
                keyp_num = np.array(self._keypoints[1])
                return np.hstack((keyp_num.reshape((-1, 1)), keyp))
            else:
                return np.array(self._keypoints[0])

        keypoints, kp_num = self._keypoints

        if return_as_list:
            keypoints_ = []

            for each_point, each_id in zip(keypoints, kp_num):
                keypoints_ps = pv.PolyData([each_point])
                keypoints_ps["entity_num"] = np.array(each_id).reshape((1))
                keypoints_.append(keypoints_ps)

            return keypoints_

        else:
            keypoints_pd = pv.PolyData(keypoints)
            keypoints_pd["entity_num"] = kp_num
            return keypoints_pd

    @property
    def _lines(self) -> List["pv.PolyData"]:
        """Cache of the lines."""
        if self._lines_cache is None:
            self._lines_cache = self._load_lines()
        return self._lines_cache

    @property
    @requires_package("pyvista")
    def lines(self) -> "pv.MultiBlock":
        """Geometry of the lines.

        Obtain the selected lines as a :class:`pyvista.MultiBlock` object.

        Returns
        -------
        pv.MultiBlock

        Examples
        --------

        >>> mapdl.geometry.lines
        MultiBlock (0x147b77e20)
          N Blocks    45
          X Bounds    -0.016, 0.016
          Y Bounds    -0.008, 0.018
          Z Bounds    -0.003, 0.015

        You can plot the entities:

        >>> mapdl.geometry.lines.plot()

        You can access the individual lines using indexing:

        >>> line0 = mapdl.geometry.lines[0]
        >>> line0
        PolyData (0x147d5b220)
          N Cells:    1
          N Points:   100
          N Strips:   0
          X Bounds:   0.000e+00, 0.000e+00
          Y Bounds:   -7.620e-03, 1.778e-02
          Z Bounds:   -3.180e-03, -3.180e-03
          N Arrays:   1

        You can use the entity name:

        >>> line1 = mapdl.geometry.lines["line 1"]
        >>> line1
        PolyData (0x147d5b220)
          N Cells:    1
          N Points:   100
          N Strips:   0
          X Bounds:   0.000e+00, 0.000e+00
          Y Bounds:   -7.620e-03, 1.778e-02
          Z Bounds:   -3.180e-03, -3.180e-03
          N Arrays:   1

        You can iterate the different elements:

        >>> points = mapdl.geometry.lines
        >>> for each_line in points:
                print(each_line)
        PolyData (0x147d5b220)
          N Cells:    1
          N Points:   100
          N Strips:   0
          X Bounds:   0.000e+00, 0.000e+00
          Y Bounds:   -7.620e-03, 1.778e-02
          Z Bounds:   -3.180e-03, -3.180e-03
          N Arrays:   1
        ...

        """
        mb = Multiblock(self._lines)
        # Setting names
        for ind, each_block in enumerate(mb):
            mapdl_index = int(each_block["entity_num"][0])
            mb.set_block_name(index=ind, name=f"line {mapdl_index}")
        return mb

    @requires_package("pyvista")
    def get_lines(
        self, return_as_list: bool = False
    ) -> Union["pv.PolyData", List["pv.PolyData"]]:
        """Obtain line geometry

        Obtain the active lines as a :class:`pyvista.PolyData` object or
        a list of :class:`pyvista.PolyData` objects.

        Parameters
        ----------
        return_as_list : bool, optional
            Whether to return the lines in a list. The default is ``False``.

        Returns
        -------
        Union[pv.PolyData, List[pv.PolyData]]

        Examples
        --------

        Return a single merged mesh.

        >>> line_mesh = mapdl.geometry.get_lines()
        >>> line_mesh
        PolyData (0x14917e740)
          N Cells:    45
          N Points:   4500
          N Strips:   0
          X Bounds:   -1.588e-02, 1.588e-02
          Y Bounds:   -7.620e-03, 1.778e-02
          Z Bounds:   -3.180e-03, 1.524e-02
          N Arrays:   1

        Return a list of lines as individual grids

        >>> lines = mapdl.geometry.get_lines(return_as_list=True)
        >>> lines
        [PolyData (0x1492ee7a...rrays:   1, PolyData (0x1491a404...rrays:   1, PolyData (0x1491068c...rrays:   1, PolyData (0x14910662...rrays:   1, PolyData (0x14910632...rrays:   1, PolyData (0x1492eeb0...rrays:   1, PolyData (0x1492eec8...rrays:   1, PolyData (0x1492eee0...rrays:   1, PolyData (0x1492eef8...rrays:   1, PolyData (0x1492ef10...rrays:   1, PolyData (0x1492ef28...rrays:   1, PolyData (0x1492ef40...rrays:   1, PolyData (0x1492ef58...rrays:   1, PolyData (0x1492ef70...rrays:   1, ...]

        """
        if return_as_list:
            return self._lines
        else:
            return merge_polydata(self._lines)

    @property
    @requires_package("pyvista")
    def areas(self) -> "pv.MultiBlock":
        """Geometry of the areas.

        Obtain the selected areas as a :class:`pyvista.MultiBlock` object.

        Returns
        -------
        pv.MultiBlock

        Examples
        --------

        >>> mapdl.geometry.areas
        MultiBlock (0x147ca7640)
          N Blocks    28
          X Bounds    -0.016, 0.016
          Y Bounds    -0.008, 0.018
          Z Bounds    -0.003, 0.015

        You can plot the entities:

        >>> mapdl.geometry.areas.plot()

        You can access the individual areas using indexing:

        >>> area0 = mapdl.geometry.areas[0]
        >>> area0
        UnstructuredGrid (0x147ca4340)
          N Cells:    10
          N Points:   18
          X Bounds:   0.000e+00, 1.588e-02
          Y Bounds:   -7.620e-03, 1.778e-02
          Z Bounds:   -3.180e-03, -3.180e-03
          N Arrays:   3

        You can use the entity name:

        >>> area1 = mapdl.geometry.areas["area 1"]
        >>> area1
        UnstructuredGrid (0x147ca4340)
          N Cells:    10
          N Points:   18
          X Bounds:   0.000e+00, 1.588e-02
          Y Bounds:   -7.620e-03, 1.778e-02
          Z Bounds:   -3.180e-03, -3.180e-03
          N Arrays:   3

        You can iterate the different elements:

        >>> points = mapdl.geometry.areas
        >>> for each_line in points:
                print(each_line)
        UnstructuredGrid (0x147ca4340)
          N Cells:    10
          N Points:   18
          X Bounds:   0.000e+00, 1.588e-02
          Y Bounds:   -7.620e-03, 1.778e-02
          Z Bounds:   -3.180e-03, -3.180e-03
          N Arrays:   3
        ...

        """
        mb = Multiblock(self.get_areas(return_as_list=True))
        # Setting names
        for ind, each_block in enumerate(mb):
            mapdl_index = int(each_block["entity_num"][0])
            mb.set_block_name(index=ind, name=f"area {mapdl_index}")
        return mb

    @requires_package("pyvista")
    def get_areas(
        self, quality: int = 1, return_as_list: Optional[bool] = False
    ) -> Union[List["pv.UnstructuredGrid"], "pv.PolyData"]:
        """Get active areas from MAPDL represented as :class:`pyvista.PolyData` or a list of :class:`pyvista.UnstructuredGrid`.

        Parameters
        ----------
        quality : int, optional
            Quality of the mesh to display.  Values are between 1 (worst)
            and10 (best).

        Returns
        -------
        pv.PolyData or List[pv.UnstructuredGrid]
            pv.PolyData grouping all meshes representing
            the active surface areas selected by ``ASEL``.  If
            ``return_as_list=True``, areas are returned as a
            list of ``pv.UnstructuredGrid``.

        Examples
        --------

        Return a single merged mesh.

        >>> area_mesh = mapdl.geometry.get_areas(quality=3)
        >>> area_mesh
        UnstructuredGrid (0x7f14add95ca0)
          N Cells:	24
          N Points:	30
          X Bounds:	-2.000e+00, 2.000e+00
          Y Bounds:	0.000e+00, 1.974e+00
          Z Bounds:	5.500e-01, 5.500e-01
          N Arrays:	4

        Return a list of areas as individual grids

        >>> areas = mapdl.geometry.get_areas(quality=3, return_as_list=True)
        >>> areas
        [UnstructuredGrid (0x7f14add95040)
          N Cells:	12
          N Points:	20
          X Bounds:	-2.000e+00, 2.000e+00
          Y Bounds:	0.000e+00, 1.974e+00
          Z Bounds:	0.000e+00, 0.000e+00
          N Arrays:	4,
        UnstructuredGrid (0x7f14add95ca0)
          N Cells:	12
          N Points:	20
          X Bounds:	-2.000e+00, 2.000e+00
          Y Bounds:	0.000e+00, 1.974e+00
          Z Bounds:	5.500e-01, 5.500e-01
          N Arrays:	4,
        ...

        """
        if not isinstance(quality, int) or (quality > 10 or quality < 1):
            raise ValueError(
                "The argument 'quality' can only be an integer between 1 and 10 (both included)."
            )

        surf = self.generate_surface(11 - quality)

        if not return_as_list:
            return surf

        entity_num = surf["entity_num"]

        areas = []
        anums = np.unique(entity_num)
        for anum in anums:
            areas.append(surf.extract_cells(entity_num == anum))

        return areas

    @supress_logging
    @run_as_prep7
    @requires_package("pyvista")
    def generate_surface(
        self,
        density: int = 4,
        amin: Optional[int] = None,
        amax: Optional[int] = None,
        ninc: Optional[int] = None,
    ) -> "pv.PolyData":
        """
        Generate an all-triangular surface of the active surfaces.

        Parameters
        ----------
        density : int, optional
            APDL smart sizing option.  Ranges from 1 (worst) to 10
            (best).

        amin : int, optional
            Minimum APDL numbered area to select.  See
            ``mapdl.anum`` for available areas.

        amax : int, optional
            Maximum APDL numbered area to select.  See
            ``mapdl.anum`` for available areas.

        ninc : int, optional
            Steps to between amin and amax.

        """
        with self._mapdl.save_selection:
            orig_anum = self.anum

            # reselect from existing selection to mimic APDL behavior
            if amin or amax:
                amax = amax or amin
                amin = amin or 1
                ninc = ninc or ""

                self._mapdl.asel("R", "AREA", vmin=amin, vmax=amax, vinc=ninc)

            ## Duplication
            # duplicate areas to avoid affecting existing areas
            # Getting the maximum area ID
            a_num = int(self._mapdl.get(entity="AREA", item1="NUM", it1num="MAXD"))
            # Setting the new areas ID starting number
            self._mapdl.numstr("AREA", a_num, mute=True)
            # Generating new areas
            self._mapdl.agen(2, "ALL", noelem=1, mute=True)

            # Getting the new maximum area ID
            a_max = int(self._mapdl.get(entity="AREA", item1="NUM", it1num="MAXD"))
            self._mapdl.asel("S", "AREA", vmin=a_num + 1, vmax=a_max, mute=True)

            # necessary to reset element/area meshing association
            self._mapdl.aatt(mute=True)

            # create a temporary etype
            etype_max = int(self._mapdl.get(entity="ETYP", item1="NUM", it1num="MAX"))
            etype_old = self._mapdl.parameters.type
            etype_tmp = etype_max + 1

            old_routine = self._mapdl.parameters.routine

            self._mapdl.et(etype_tmp, "MESH200", 6, mute=True)
            self._mapdl.shpp("off", mute=True)
            self._mapdl.smrtsize(density, mute=True)
            self._mapdl.type(etype_tmp, mute=True)

            if old_routine != "PREP7":
                self._mapdl.prep7(mute=True)

            # Mesh and get the number of elements per area
            resp = self._mapdl.amesh("all")
            elements_per_area = self.get_elements_per_area()

            self._mapdl.esla("S")
            grid = self._mapdl.mesh._grid.linear_copy()
            pd = pv.PolyData(grid.points, grid.cells)

            # pd['ansys_node_num'] = grid['ansys_node_num']
            # pd['vtkOriginalPointIds'] = grid['vtkOriginalPointIds']
            # pd.clean(inplace=True)  # OPTIONAL

            # delete all temporary meshes and clean up settings
            self._mapdl.aclear("ALL", mute=True)
            self._mapdl.adele("ALL", kswp=1, mute=True)
            self._mapdl.numstr("AREA", 1, mute=True)
            self._mapdl.type(etype_old, mute=True)
            self._mapdl.etdele(etype_tmp, mute=True)
            self._mapdl.shpp("ON", mute=True)
            self._mapdl.smrtsize("OFF", mute=True)

        # store the area number used for each element
        entity_num = np.empty(grid.n_cells, dtype=np.int32)
        if grid and len(elements_per_area) != 0:
            # add anum info
            i = 0
            for index, (anum, nelem) in enumerate(elements_per_area):
                # have to use original area numbering here as the
                # duplicated areas numbers are inaccurate
                entity_num[i : i + nelem] = orig_anum[index]
                i += nelem
        else:
            entity_num[:] = 0

        pd["entity_num"] = entity_num
        return pd

    @property
    def n_volu(self) -> int:
        """
        Number of volumes currently selected.

        Examples
        --------
        >>> mapdl.n_volu
        1
        """
        return self._item_count("VOLU")

    @property
    def n_area(self) -> int:
        """
        Number of areas currently selected.

        Examples
        --------
        >>> mapdl.n_area
        1
        """
        return self._item_count("AREA")

    @property
    def n_line(self) -> int:
        """
        Number of lines currently selected.

        Examples
        --------
        >>> mapdl.n_line
        1
        """
        return self._item_count("LINE")

    @property
    def n_keypoint(self) -> int:
        """
        Number of keypoints currently selected.

        Examples
        --------
        >>> mapdl.n_keypoint
        1
        """
        return self._item_count("KP")

    @supress_logging
    def _item_count(self, entity: str) -> int:
        """Return item count for a given entity."""
        return int(self._mapdl.get(entity=entity, item1="COUNT"))

    @property
    def knum(self) -> NDArray[np.int32]:
        """
        Array of keypoint numbers.

        Examples
        --------
        >>> mapdl.block(0, 1, 0, 1, 0, 1)
        >>> mapdl.knum
        array([1, 2, 3, 4, 5, 6, 7, 8], dtype=int32)
        """
        if self._mapdl.geometry.n_keypoint == 0:
            return np.array([], dtype=np.int32)

        return self._mapdl.get_array("KP", item1="KLIST").astype(np.int32)

    @property
    def lnum(self) -> NDArray[np.int32]:
        """Array of line numbers.

        Examples
        --------
        >>> mapdl.block(0, 1, 0, 1, 0, 1)
        >>> mapdl.lnum
        array([ 1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12], dtype=int32)
        """
        # For clean exit when there is no lines.
        if self._mapdl.geometry.n_line == 0:
            return np.array([], dtype=np.int32)

        # this (weirdly) sometimes fails
        for _ in range(5):
            lnum = self._mapdl.get_array("LINES", item1="LLIST")
            if lnum.size == self.n_line:
                break
        return lnum.astype(np.int32)

    @property
    def anum(self) -> NDArray[np.int32]:
        """Array of area numbers.

        Examples
        --------
        >>> mapdl.block(0, 1, 0, 1, 0, 1)
        >>> mapdl.anum
        array([1, 2, 3, 4, 5, 6], dtype=int32)
        """
        # Clean exit
        if self._mapdl.geometry.n_area == 0:
            return np.array([], dtype=np.int32)

        return self._mapdl.get_array("AREA", item1="ALIST").astype(np.int32)

    @property
    def vnum(self) -> NDArray[np.int32]:
        """Array of volume numbers.

        Examples
        --------
        >>> mapdl.block(0, 1, 0, 1, 0, 1)
        >>> mapdl.vnum
        array([1], dtype=int32)
        """
        if self._mapdl.geometry.n_volu == 0:
            return np.array([], dtype=np.int32)

        return self._mapdl.get_array("VOLU", item1="VLIST").astype(np.int32)

    @supress_logging
    @requires_package("pyvista")
    @requires_package("pyiges")
    def _load_lines(self) -> List["pv.PolyData"]:
        """Load lines from MAPDL using IGES"""
        # ignore volumes
        with self._mapdl.save_selection:
            self._mapdl.ksel("ALL", mute=True)
            self._mapdl.lsel("ALL", mute=True)
            self._mapdl.asel("ALL", mute=True)
            self._mapdl.vsel("NONE", mute=True)

            iges = self._load_iges()

        selected_lnum = self.lnum
        lines = []
        entity_nums = []
        for bspline in iges.bsplines():
            # allow only 10001 as others appear to be construction entities
            if bspline.d["status_number"] in [1, 10001]:
                entity_num = int(bspline.d["entity_subs_num"])
                if entity_num not in entity_nums and entity_num in selected_lnum:
                    entity_nums.append(entity_num)
                    line = bspline.to_vtk()
                    line.cell_data["entity_num"] = entity_num
                    lines.append(line)

        entities = iges.lines() + iges.circular_arcs()
        for line in entities:
            if line.d["status_number"] == 1:
                entity_num = int(line.d["entity_subs_num"])
                if entity_num not in entity_nums and entity_num in selected_lnum:
                    entity_nums.append(entity_num)
                    line = line.to_vtk(resolution=1)
                    line.cell_data["entity_num"] = entity_num
                    lines.append(line)

        lines_ = []
        for line in lines:
            lines_.append(line)

        return lines_

    @requires_package("pyiges")
    def _load_keypoints(self) -> Tuple[NDArray, NDArray]:
        """Load keypoints from MAPDL using IGES."""
        # write only keypoints
        with self._mapdl.save_selection:
            self._mapdl.vsel("NONE", mute=True)
            self._mapdl.asel("NONE", mute=True)
            self._mapdl.lsel("NONE", mute=True)
            iges = self._load_iges()

        keypoints = []
        kp_num = []
        for kp in iges.points():
            keypoints.append([kp.x, kp.y, kp.z])
            kp_num.append(int(kp.d["entity_subs_num"]))

        return keypoints, kp_num

    def __str__(self) -> str:
        """Current geometry info"""
        info = "MAPDL Selected Geometry\n"
        info += "Keypoints:  %d\n" % self.n_keypoint
        info += "Lines:      %d\n" % self.n_line
        info += "Areas:      %d\n" % self.n_area
        info += "Volumes:    %d\n" % self.n_volu
        return info

    def keypoint_select(
        self,
        items: Optional[Sequence[int]] = None,
        sel_type: VALID_SELECTION_TYPE_TP = "S",
        return_selected: bool = False,
    ) -> Optional[NDArray[np.int32]]:
        """Select keypoints using a sequence of items.

        Parameters
        ----------
        items : sequence or None
            List, range, or sequence of integers of the keypoints you want
            to select.  If ``None`` or ``'NONE'``, no keypoints will be
            selected.  If 'ALL', selects all keypoints.

        sel_type : str, optional
            Selection type.  May be one of the following:

            * ``'S'``: Select a new set (default)
            * ``'R'``: Reselect a set from the current set.
            * ``'A'``: Additionally select a set and extend the current set.
            * ``'U'``: Unselect a set from the current set.

        return_selected : bool, optional
            Return the keypoint numbers selected.  Optional, and can be
            disabled for performance.  Default ``False``.

        Returns
        -------
        np.array
            Numpy array of keypoint numbers if ``return_selected=True``.

        Examples
        --------
        Create a new selection of keypoints [1, 5, 10]

        >>> mapdl.geometry.keypoint_select([1, 5, 10])

        Create a new selection of keypoints from 1 through 20

        >>> mapdl.geometry.keypoint_select(range(1, 21))

        Unselect keypoints 1 through 20

        >>> mapdl.geometry.keypoint_select(range(1, 21), sel_type='U')

        Append to an existing selection of keypoints

        >>> mapdl.geometry.keypoint_select([1, 2, 3], sel_type='A')

        Reselect from the existing selection of keypoints

        >>> mapdl.geometry.keypoint_select([3, 4, 5], sel_type='R')

        Select no keypoints

        >>> mapdl.geometry.keypoint_select(None)

        Select all keypoints

        >>> mapdl.geometry.keypoint_select('ALL')

        """
        if isinstance(items, str):
            items = items.upper()

        # special cases
        if items is None or items == "NONE":
            self._mapdl.ksel("NONE")
            return

        if items == "ALL":
            self._mapdl.ksel("ALL")
            if return_selected:
                return self.knum
            return

        self._select_items(items, "KP", sel_type)

        if return_selected:
            return self.knum

    def line_select(
        self,
        items: Optional[Sequence[int]],
        sel_type: VALID_SELECTION_TYPE_TP = "S",
        return_selected: bool = False,
    ) -> Optional[NDArray[np.int32]]:
        """Select lines using a sequence of items.

        Parameters
        ----------
        items : sequence or None
            List, range, or sequence of integers of the lines you want
            to select.  If ``None`` or ``'NONE'``, no lines will be
            selected.  If 'ALL', selects all lines.

        sel_type : str, optional
            Selection type.  May be one of the following:

            * ``'S'``: Select a new set (default)
            * ``'R'``: Reselect a set from the current set.
            * ``'A'``: Additionally select a set and extend the current set.
            * ``'U'``: Unselect a set from the current set.

        return_selected : bool, optional
            Return the line numbers selected.  Optional, and can be
            disabled for performance.  Default ``False``.

        Returns
        -------
        np.array
            Numpy array of keypoint numbers if ``return_selected=True``.

        Examples
        --------
        Create a new selection of lines [1, 5, 10]

        >>> mapdl.geometry.line_select([1, 5, 10])

        Create a new selection of lines from 1 through 20

        >>> mapdl.geometry.line_select(range(1, 21))

        Unselect lines 1 through 20

        >>> mapdl.geometry.line_select(range(1, 21), sel_type='U')

        Append to an existing selection of lines

        >>> mapdl.geometry.line_select([1, 2, 3], sel_type='A')

        Reselect from the existing selection of lines

        >>> mapdl.geometry.line_select([3, 4, 5], sel_type='R')

        Select no lines

        >>> mapdl.geometry.line_select(None)

        Select all lines

        >>> mapdl.geometry.line_select('ALL')

        """
        if isinstance(items, str):
            items = items.upper()

        # special cases
        if items is None or items == "NONE":
            self._mapdl.lsel("NONE")
            return

        if items == "ALL":
            self._mapdl.lsel("ALL")
            if return_selected:
                return self.lnum
            return

        self._select_items(items, "LINE", sel_type)

        if return_selected:
            return self.lnum

    def area_select(
        self,
        items: Optional[Sequence[int]],
        sel_type: VALID_SELECTION_TYPE_TP = "S",
        return_selected: bool = False,
    ) -> Optional[NDArray[np.int32]]:
        """Select areas using a sequence of items.

        Parameters
        ----------
        items : sequence, str, None
            List, range, or sequence of integers of the areas you want
            to select.  If ``None`` or ``'NONE'``, no areas will be
            selected.  If 'ALL', selects all areas.

        sel_type : str, optional
            Selection type.  May be one of the following:

            * ``'S'``: Select a new set (default)
            * ``'R'``: Reselect a set from the current set.
            * ``'A'``: Additionally select a set and extend the current set.
            * ``'U'``: Unselect a set from the current set.

        return_selected : bool, optional
            Return the area numbers selected.  Optional, and can be
            disabled for performance.  Default ``False``.

        Returns
        -------
        np.array
            Numpy array of keypoint numbers if ``return_selected=True``.

        Examples
        --------
        Create a new selection of areas [1, 5, 10]

        >>> mapdl.geometry.area_select([1, 5, 10])

        Create a new selection of areas from 1 through 20

        >>> mapdl.geometry.area_select(range(1, 21))

        Unselect areas 1 through 20

        >>> mapdl.geometry.area_select(range(1, 21), sel_type='U')

        Append to an existing selection of areas

        >>> mapdl.geometry.area_select([1, 2, 3], sel_type='A')

        Reselect from the existing selection of areas

        >>> mapdl.geometry.area_select([3, 4, 5], sel_type='R')

        Select no areas

        >>> mapdl.geometry.area_select(None)

        Select all areas

        >>> mapdl.geometry.area_select('ALL')

        """
        if isinstance(items, str):
            items = items.upper()

        # special cases
        if items is None or items == "NONE":
            self._mapdl.asel("NONE")
            return

        if items == "ALL":
            self._mapdl.asel("ALL")
            if return_selected:
                return self.anum
            return

        self._select_items(items, "AREA", sel_type)

        if return_selected:
            return self.anum

    @property
    @requires_package("pyvista")
    def volumes(self) -> "pv.MultiBlock":
        """Obtain the volumes geometry

        Obtain the selected volumes as a :class:`pyvista.MultiBlock` object.

        Returns
        -------
        pv.MultiBlock

        Examples
        --------

        >>> mapdl.geometry.volumes
        MultiBlock (0x147ca4100)
          N Blocks    6
          X Bounds    -0.016, 0.016
          Y Bounds    -0.008, 0.018
          Z Bounds    -0.003, 0.015

        You can plot the entities:

        >>> mapdl.geometry.volumes.plot()

        You can access the individual volumes using indexing:

        >>> volume0 = mapdl.geometry.volumes[0]
        >>> volume0
        UnstructuredGrid (0x149107340)
          N Cells:    34
          N Points:   36
          X Bounds:   0.000e+00, 1.588e-02
          Y Bounds:   -7.620e-03, 1.778e-02
          Z Bounds:   -3.180e-03, 0.000e+00
          N Arrays:   3

        You can use the entity name:

        >>> volume1 = mapdl.geometry.volumes["volume 1"]
        >>> volume1
        UnstructuredGrid (0x149107340)
          N Cells:    34
          N Points:   36
          X Bounds:   0.000e+00, 1.588e-02
          Y Bounds:   -7.620e-03, 1.778e-02
          Z Bounds:   -3.180e-03, 0.000e+00
          N Arrays:   3

        You can iterate the different elements:

        >>> points = mapdl.geometry.volumes
        >>> for each_line in points:
                print(each_line)
        UnstructuredGrid (0x149107340)
          N Cells:    34
          N Points:   36
          X Bounds:   0.000e+00, 1.588e-02
          Y Bounds:   -7.620e-03, 1.778e-02
          Z Bounds:   -3.180e-03, 0.000e+00
          N Arrays:   3
        ...

        """
        mb = Multiblock(self.get_volumes(return_as_list=True))
        # Setting names
        # Because the volume mapping is made through the areas, the default
        # "entity_num" field is not applicable here.
        for ind, each_block in enumerate(mb):
            mapdl_index = int(each_block.entity_num)
            mb.set_block_name(index=ind, name=f"volume {mapdl_index}")
        return mb

    @requires_package("pyvista")
    def get_volumes(
        self, return_as_list: bool = False, quality: int = 4
    ) -> Union[List["pv.PolyData"], "pv.PolyData"]:
        """Get active volumes from MAPDL represented as a :class:`pyvista.PolyData` object
        or a list of :class:`pyvista.UnstructuredGrid` objects.

        Parameters
        ----------
        quality : int, optional
            Quality of the mesh to display.  Values are 1 (worst)
            to 10 (best).

        Returns
        -------
        pv.PolyData or List[pv.UnstructuredGrid]
            pv.PolyData grouping all meshes representing
            the active surface volumes selected by ``VSEL``.  If
            ``return_as_list=True``, volumes are returned as a
            list of ``pv.UnstructuredGrid``.

        Examples
        --------

        Return a single merged mesh.

        >>> volume_mesh = mapdl.geometry.get_volumes(quality=3)
        >>> volume_mesh
        UnstructuredGrid (0x7f14add95ca0)
          N Cells:	24
          N Points:	30
          X Bounds:	-2.000e+00, 2.000e+00
          Y Bounds:	0.000e+00, 1.974e+00
          Z Bounds:	5.500e-01, 5.500e-01
          N Arrays:	4

        Return a list of volumes as individual grids.

        >>> volumes = mapdl.geometry.get_volumes(quality=3, return_as_list=True)
        >>> volumes
        [UnstructuredGrid (0x7f14add95040)
          N Cells:	12
          N Points:	20
          X Bounds:	-2.000e+00, 2.000e+00
          Y Bounds:	0.000e+00, 1.974e+00
          Z Bounds:	0.000e+00, 0.000e+00
          N Arrays:	4,
        UnstructuredGrid (0x7f14add95ca0)
          N Cells:	12
          N Points:	20
          X Bounds:	-2.000e+00, 2.000e+00
          Y Bounds:	0.000e+00, 1.974e+00
          Z Bounds:	5.500e-01, 5.500e-01
          N Arrays:	4,
        ...

        """
        quality = int(quality)
        if quality > 10:
            raise ValueError("``quality`` parameter must be a value between 0 and 10")

        volumes_ = []
        surf = self.generate_surface(11 - quality)

        if not return_as_list:
            return surf

        # Cache current selection
        with self._mapdl.save_selection:
            area_num = surf["entity_num"].astype(int)

            for each_volu in self.vnum:
                self._mapdl.vsel("S", vmin=each_volu)
                self._mapdl.aslv("S")
                unstruct = surf.extract_cells(np.in1d(area_num, self.anum))
                unstruct.entity_num = int(each_volu)
                volumes_.append(unstruct)

        return volumes_

    def volume_select(
        self,
        items: Optional[Union[str, Sequence[int]]],
        sel_type: VALID_SELECTION_TYPE_TP = "S",
        return_selected: bool = False,
    ) -> Optional[int]:
        """Select volumes using a sequence of items.

        Parameters
        ----------
        items : sequence, str, or None
            List, range, or sequence of integers of the volumes you want
            to select.  If ``None`` or ``'NONE'``, no volumes will be
            selected.  If 'ALL', selects all volumes.

        sel_type : str, optional
            Selection type.  May be one of the following:

            * ``'S'``: Select a new set (default)
            * ``'R'``: Reselect a set from the current set.
            * ``'A'``: Additionally select a set and extend the current set.
            * ``'U'``: Unselect a set from the current set.

        return_selected : bool, optional
            Return the volume numbers selected.  Optional, and can be
            disabled for performance.  Default ``False``.

        Returns
        -------
        np.array
            Numpy array of keypoint numbers if ``return_selected=True``.

        Examples
        --------
        Create a new selection of volumes [1, 5, 10]

        >>> mapdl.geometry.volume_select([1, 5, 10])

        Create a new selection of volumes from 1 through 20

        >>> mapdl.geometry.volume_select(range(1, 21))

        Unselect volumes 1 through 20

        >>> mapdl.geometry.volume_select(range(1, 21), sel_type='U')

        Append to an existing selection of volumes

        >>> mapdl.geometry.volume_select([1, 2, 3], sel_type='A')

        Reselect from the existing selection of volumes

        >>> mapdl.geometry.volume_select([3, 4, 5], sel_type='R')

        Select no volumes

        >>> mapdl.geometry.volume_select(None)

        Select all volumes

        >>> mapdl.geometry.volume_select('ALL')

        """
        if isinstance(items, str):
            items = items.upper()

        # special cases
        if items is None or items == "NONE":
            self._mapdl.vsel("NONE")
            return

        if items == "ALL":
            self._mapdl.vsel("ALL")
            if return_selected:
                return self.vnum
            return

        self._select_items(items, "VOLU", sel_type)

        if return_selected:
            return self.vnum

    def _select_items(
        self,
        items: Sequence[int],
        item_type: VALID_SELECTION_ENTITY_TP,
        sel_type: VALID_SELECTION_TYPE_TP,
    ) -> None:
        """Select items using FLST.

        Parameters
        ----------
        items : sequence
            Sequence of items.

        item_type : str
            Item lookup type.  One of:

               * 'NODE' : node numbers
               * 'ELEM' : element numbers
               * 'KP' : keypoint numbers
               * 'LINE' : line numbers
               * 'AREA' : area numbers
               * 'VOLU' : volume numbers
               * 'TRACE' : trace points
               * 'COORD' : coordinate locations

        sel_type : str, optional
            Selection type.  Must be one of the following:

            * ``'S'``: Select a new set (default)
            * ``'R'``: Reselect a set from the current set.
            * ``'A'``: Additionally select a set and extend the current set.
            * ``'U'``: Unselect a set from the current set.

        """
        if item_type not in FLST_LOOKUP:
            raise KeyError(f'Invalid ``item_type`` "{item_type}"')

        sel_type = sel_type.upper()
        if sel_type not in VALID_SELECTION_TYPE:
            raise ValueError(
                f'Invalid ``sel_type`` "{sel_type}"\n\n'
                f"Use one of the following:\n{VALID_TYPE_MSG}"
            )

        # convert to a flat array as it's easier for type checking
        items = np.asarray(items)
        if not np.issubdtype(items.dtype, np.number):
            raise TypeError("Item numbers must be a numeric type")
        items = items.ravel().astype(np.int_, copy=False)

        # We can use list or arrays for vmin
        if item_type == "NODE":
            self._mapdl.nsel(sel_type, vmin=items, return_mapdl_output=True)
        elif item_type == "ELEM":
            self._mapdl.esel(sel_type, vmin=items, return_mapdl_output=True)
        elif item_type == "KP":
            self._mapdl.ksel(sel_type, vmin=items, return_mapdl_output=True)
        elif item_type == "LINE":
            self._mapdl.lsel(sel_type, vmin=items, return_mapdl_output=True)
        elif item_type == "AREA":
            self._mapdl.asel(sel_type, vmin=items, return_mapdl_output=True)
        elif item_type == "VOLU":
            self._mapdl.vsel(sel_type, vmin=items, return_mapdl_output=True)
        else:
            raise ValueError(f'Unable to select "{item_type}"')

    def get_elements_per_area(self) -> NDArray[np.int32]:
        """Get the number of elements meshed for each area.

        Returns
        -------
        np.ndarray
            An array with the area id for the first column, and the number of
            elements per each area on the second column.

        """
        anum = self.anum.ravel()

        elem_per_areas = self._mapdl.get_array("area", "", "ATTR", "NELM")
        elem_per_areas = elem_per_areas[anum - 1].ravel()

        return np.vstack((anum, elem_per_areas)).T.astype(np.int32)


class LegacyGeometry(Geometry):
    """Legacy Pythonic representation of the MAPDL CAD geometry.

    This class contains advanced methods for extending geometry building and
    selection within MAPDL.
    """

    def __init__(self, mapdl: Mapdl):
        """Legacy geometry manager

        Class to help to manage geometry representations in an
        :class:`Mapdl instance <ansys.mapdl.core.Mapdl>` instance.

        Parameters
        ----------
        mapdl : ansys.mapdl.core.Mapdl
            Mapdl instance which this class references to.
        """
        super().__init__(mapdl)

    def keypoints(self) -> np.array:  # type: ignore
        """Keypoint coordinates."""
        return super().get_keypoints(return_as_array=True)

    @requires_package("pyvista")
    def lines(self) -> "pv.PolyData":
        """Active lines as a ``pyvista.PolyData`` object."""
        return super().get_lines()  # type: ignore

    @requires_package("pyvista")
    def areas(
        self, quality=1, merge=False
    ) -> Union["pv.PolyData", List["pv.UnstructuredGrid"]]:
        """List of areas from MAPDL represented as a ``pyvista.PolyData`` object.

        Parameters
        ----------
        quality : int, optional
            Quality of the mesh to display.  Values are 1 (worst)
            to 10 (best).

        merge : bool, optional
            Whether to merge areas into a single mesh. The default
            is ``False``, in which case a list of areas is returned.  When ``True``,
            the output is a single mesh.

        Returns
        -------
        list of pyvista.UnstructuredGrid
            List of :class:`pyvista.UnstructuredGrid <pyvista.UnstructuredGrid>` meshes representing
            the active surface areas selected by ``ASEL``.  If
            ``merge=True``, areas are returned as a single merged
            :class:`pyvista.UnstructuredGrid <pyvista.UnstructuredGrid>`.

        Examples
        --------
        Return a list of areas as individual grids.

        >>> areas = mapdl.areas(quality=3)
        >>> areab
        [UnstructuredGrid (0x7f14add95040)
          N Cells:	12
          N Points:	20
          X Bounds:	-2.000e+00, 2.000e+00
          Y Bounds:	0.000e+00, 1.974e+00
          Z Bounds:	0.000e+00, 0.000e+00
          N Arrays:	4,
        UnstructuredGrid (0x7f14add95ca0)
          N Cells:	12
          N Points:	20
          X Bounds:	-2.000e+00, 2.000e+00
          Y Bounds:	0.000e+00, 1.974e+00
          Z Bounds:	5.500e-01, 5.500e-01
          N Arrays:	4,
        ...

        Return a single merged mesh.

        >>> area_mesh = mapdl.areas(quality=3)
        >>> area_mesh
        UnstructuredGrid (0x7f14add95ca0)
          N Cells:	24
          N Points:	30
          X Bounds:	-2.000e+00, 2.000e+00
          Y Bounds:	0.000e+00, 1.974e+00
          Z Bounds:	5.500e-01, 5.500e-01
          N Arrays:	4
        """
        return super().get_areas(quality=quality, return_as_list=not merge)



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class ProcessControls:
    def wait(self, dtime="", **kwargs):
        """APDL Command: /WAIT

        Causes a delay before the reading of the next command.

        Parameters
        ----------
        dtime
            Time delay (in seconds). Maximum time delay is 59 seconds.

        Notes
        -----
        You should consider using ``time.sleep(dtime)``

        The command following the /WAIT will not be processed until the
        specified wait time increment has elapsed.  Useful when reading from a
        prepared input file to cause a pause, for example, after a display
        command so that the display can be reviewed for a period of time.
        Another "wait" feature is available via the ``*ASK`` command.

        This command is valid in any processor.
        """
        command = f"/WAIT,{dtime}"
        return self.run(command, **kwargs)



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class GeneralRadiation:
    def stef(self, value="", **kwargs):
        """Specifies Stefan-Boltzmann radiation constant.

        APDL Command: STEF

        Parameters
        ----------
        value
            Stefan-Boltzmann constant (defaults to 0.119E-10 Btu/hr/in2/ °R4).

        Notes
        -----
        You can use this command in the general preprocessor (PREP7) and in the
        Solution processor to specify the Stefan-Boltzmann constant in analyses
        using the radiation matrix method or the radiosity solver to model
        radiation.
        """
        command = f"STEF,{value}"
        return self.run(command, **kwargs)

    def toffst(self, value="", **kwargs):
        """Specifies the temperature offset from absolute zero to zero.

        APDL Command: TOFFST

        Parameters
        ----------
        value
            Degrees between absolute zero and zero of temperature system used
            (should be positive).

        Notes
        -----
        Specifies the difference (in degrees) between absolute zero and the
        zero of the temperature system used.  Absolute temperature values are
        required in evaluating certain expressions, such as for creep,
        swelling, radiation heat transfer, MASS71, etc.  (The offset
        temperature is not used in evaluating emissivity.) Examples are 460°
        for the Fahrenheit system and 273° for the Celsius system.  The offset
        temperature is internally included in the element calculations and does
        not affect the temperature input or output.  If used in SOLUTION, this
        command is valid only within the first load step.

        This command is also valid in PREP7.
        """
        command = f"TOFFST,{value}"
        return self.run(command, **kwargs)



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class LoadCase:
    def lcabs(self, lcno="", kabs="", **kwargs):
        """Specifies absolute values for load case operations.

        APDL Command: LCABS

        Parameters
        ----------
        lcno
            Load case pointer number.  If ALL, apply to all selected load cases
            [LCSEL].

        kabs
            Absolute value key:

            0
                Use algebraic values of load case LCNO in operations.

            1
                Use absolute values of load case LCNO in operations.

        Notes
        -----
        Causes absolute values to be used in the load case operations [LCASE or
        LCOPER].  Absolute values are taken prior to assigning a load case
        factor [LCFACT] and are applied only to defined load cases [LCDEF].
        """
        command = f"LCABS,{lcno},{kabs}"
        return self.run(command, **kwargs)

    def lcase(self, lcno="", **kwargs):
        """Reads a load case into the database.

        APDL Command: LCASE

        Parameters
        ----------
        lcno
            Load case pointer number [LCDEF,STAT].  Defaults to 1.

        Notes
        -----
        Reads a load case into the database.  Load cases are created as
        described on the LCDEF or LCWRITE commands.  The results portion of the
        database and the applied forces and displacements are cleared before
        reading the data in.  Absolute values [LCABS] and scale factors
        [LCFACT] can be applied during the read operation.
        """
        command = f"LCASE,{lcno}"
        return self.run(command, **kwargs)

    def lcdef(self, lcno="", lstep="", sbstep="", kimg="", **kwargs):
        """Creates a load case from a set of results on a results file.

        APDL Command: LCDEF

        Parameters
        ----------
        lcno
            Arbitrary pointer number (1-99) to be assigned to the load case
            specified by LSTEP, SBSTEP and by the FILE command.  Defaults to 1
            + previous value.

        lstep
            Load step number to be defined as the load case.  Defaults to one.

        sbstep
            Substep number.  Defaults to the last substep of the load step.

        kimg
            Used only with results from complex analyses:

            0
                Use real part of complex solution

            1
                Use imaginary part.

        Notes
        -----
        Creates a load case by establishing a pointer to a set of results on a
        results file (written during the ANSYS solution phase).  This pointer
        (LCNO) can then be used on the LCASE or LCOPER commands to read the
        load case data into the database.

        Issue LCDEF,ERASE to delete all load case pointers (and all load case
        files, if any).  Issue LCDEF,LCNO,ERASE to delete only the specific
        load case pointer LCNO (and its file, if any).  With the ERASE options,
        all pointers are deleted; however only files with the default extension
        [LCWRITE] are deleted.  Issue LCDEF,STAT for status of all selected
        load cases [LCSEL], or LCDEF,STAT,ALL for status of all load cases.
        The STAT command may be used to list all load cases.  See also LCFILE
        to establish a pointer to a set of results on a load case file (written
        by LCWRITE). Harmonic element data read from a result file load case is
        stored at the zero-degree position.
        """
        command = f"LCDEF,{lcno},{lstep},{sbstep},{kimg}"
        return self.run(command, **kwargs)

    def lcfact(self, lcno="", fact="", **kwargs):
        """Defines scale factors for load case operations.

        APDL Command: LCFACT

        Parameters
        ----------
        lcno
            Load case pointer number.  If ALL, apply to all selected load cases
            [LCSEL].

        fact
            Scale factor applied to load case LCNO.  Blank defaults to 1.0.

        Notes
        -----
        Defines scale factors to be used in the load case operations [LCASE or
        LCOPER].  Scale factors are applied after an absolute value operation
        [LCABS] and are applied only to defined load cases [LCDEF].
        """
        command = f"LCFACT,{lcno},{fact}"
        return self.run(command, **kwargs)

    def lcfile(self, lcno="", fname="", ext="", **kwargs):
        """Creates a load case from an existing load case file.

        APDL Command: LCFILE

        Parameters
        ----------
        lcno
            Arbitrary (1-99) pointer number assigned to this load case.

        fname
            File name and directory path (248 characters maximum, including the
            characters needed for the directory path).  An unspecified
            directory path defaults to the working directory; in this case, you
            can use all 248 characters for the file name.

        ext
            Filename extension (eight-character maximum).

        Notes
        -----
        Creates a load case by establishing a pointer to an existing load case
        file [LCWRITE].  This pointer (LCNO) can then be used on the LCASE or
        LCOPER commands to read the load case data into the database.  This
        command is typically used to reestablish load case pointers in a new
        ANSYS session (pointers are not saved on the database file), or when
        more than one pointer to a single load case is desired.  See the LCDEF
        command for status and erase operations.  See also LCDEF to establish a
        pointer to a set of results on a results file (written during the ANSYS
        solution phase).
        """
        command = f"LCFILE,{lcno},{fname},{ext}"
        return self.run(command, **kwargs)

    def lcoper(self, oper="", lcase1="", oper2="", lcase2="", **kwargs):
        """Performs load case operations.

        APDL Command: LCOPER

        Parameters
        ----------
        oper
            Valid operations are:

            ZERO
                Zero results portion of database (LCASE1 ignored).

            SQUA
                Square database values (LCASE1 ignored).

            SQRT
                Square root of database (absolute) values (LCASE1 ignored).

            LPRIN
                Recalculate line element principal stresses (LCASE1 ignored).  Stresses are as
                shown for the NMISC items of the ETABLE command for the
                specific line element type.

            ADD
                Add LCASE1 to database values.

            SUB
                Subtract LCASE1 from database values.

            SRSS
                Square root of the sum of the squares of database and LCASE1.

            MIN
                Compare and save in database the algebraic minimum of database and LCASE1.

            MAX
                Compare and save in database the algebraic maximum of database and LCASE1.

            ABMN
                Compare and save in database the absolute minimum of database and LCASE1 (based
                on magnitudes, then apply the corresponding sign).

            ABMX
                Compare and save in database the absolute maximum of database and LCASE1 (based
                on magnitudes, then apply the corresponding sign).

        lcase1
            First load case in the operation (if any).  See LCNO of the LCDEF
            command. If ALL, repeat operations using all selected load cases
            [LCSEL].

        oper2
            Valid operations are:

            MULT
                Multiplication: ``LCASE1*LCASE2``

            CPXMAX
                This option does a phase angle sweep to calculate the maximum of derived
                stresses and equivalent strain for a complex solution
                where LCASE1 is the real part and LCASE2 is the imaginary
                part. The Oper field is not applicable with  this option.
                Also, the LCABS and SUMTYPE commands have no effect on
                this option. The value of S3 will be a minimum.   This
                option does not apply to derived displacement amplitude
                (USUM). Load case writing (LCWRITE) is not supported. See
                POST1 and POST26 – Complex Results Postprocessing in the
                Mechanical APDL Theory Reference for more information.

        lcase2
            Second load case.  Used only with Oper2 operations.

        Notes
        -----
        LCOPER operates on the database and one or two load cases according to:

        Database = Database Oper (LCASE1 Oper2 LCASE2)

        where operations Oper and Oper2 are as described above.  Absolute
        values and scale factors may be applied to the load cases before the
        operations [LCABS, LCFACT].  If LCASE1 is not specified, only operation
        Oper is performed on the current database.  If LCASE2 is specified,
        operation Oper2 will be performed before operation Oper.  If LCASE2 is
        not specified, operation Oper2 is ignored.  Solution items not
        contained [OUTRES] in either the database or the applicable load cases
        will result in a null item during a load case operation.  Harmonic
        element data read from a result file load case are processed at zero
        degrees.  All load case combinations are performed in the solution
        coordinate system, and the data resulting from load case combinations
        are stored in the solution coordinate system.  The resultant data are
        then transformed to the active results coordinate system [RSYS] when
        listed or displayed. Except in the cases of Oper = LPRIN, ADD, or SUB,
        you must use RSYS,SOLU to list or display, and in the case of layered
        elements, the layer (LAYER) must also be specified.

        Use the FORCE command prior to any combination operation to correctly
        combine the requested force type.

        If Oper2=CPXMAX, the derived stresses and strain calculation do not
        apply to line elements.
        """
        command = f"LCOPER,{oper},{lcase1},{oper2},{lcase2}"
        return self.run(command, **kwargs)

    def lcsel(self, type_="", lcmin="", lcmax="", lcinc="", **kwargs):
        """Selects a subset of load cases.

        APDL Command: LCSEL

        Parameters
        ----------
        type\\_
            Label identifying the type of select:

            S
                Select a new set.

            R
                Reselect a set from the current set.

            A
                Additionally select a set and extend the current set.

            U
                Unselect a set from the current set.

            ALL
                Restore the full set.

            NONE
                Unselect the full set.

            INVE
                Invert the current set (selected becomes unselected and vice versa).

            STAT
                Display the current select status.

        lcmin
            Minimum value of load case pointer range.

        lcmax
            Maximum value of load case pointer range.  LCMAX defaults to LCMIN.

        lcinc
            Value increment within range.  Defaults to 1.  LCINC cannot be
            negative.

        Notes
        -----
        Selects a subset of load cases for other operations.  For example, to
        select a new set of load cases based on load cases 1 through 7, use
        LCSEL,S,1,7.  The subset is used when the ALL label is entered (or
        implied) on other commands, such as LCFACT, LCABS, LCOPER, etc.  Load
        cases are flagged as selected and unselected; no load case pointers
        [LCDEF, LCWRITE, LCFILE] are actually deleted from the database.
        """
        command = f"LCSEL,{type_},{lcmin},{lcmax},{lcinc}"
        return self.run(command, **kwargs)

    def lcsum(self, lab="", **kwargs):
        """Specifies whether to process non-summable items in load case

        APDL Command: LCSUM
        operations.

        Parameters
        ----------
        lab
            Combination option

            (blank)
                Only combine summable items [default].

            ALL
                Combine all items including non summable items.

        Notes
        -----
        Allows non-summable items (e.g. plastic strains) to be included in load
        combinations.  Issue LCSUM,ALL before the first load case operation
        (LCXX command).  May also be used to include nonsummable items in the
        appending of a results file (RAPPND command).
        """
        command = f"LCSUM,{lab}"
        return self.run(command, **kwargs)

    def lcwrite(self, lcno="", fname="", ext="", **kwargs):
        """Creates a load case by writing results to a load case file.

        APDL Command: LCWRITE

        Parameters
        ----------
        lcno
            Arbitrary pointer number (1-99) to be assigned to this load case.

        fname
            File name and directory path (248 characters maximum, including the
            characters needed for the directory path).  An unspecified
            directory path defaults to the working directory; in this case, you
            can use all 248 characters for the file name.

        ext
            Filename extension (eight-character maximum).

        Notes
        -----
        Creates a load case by writing the results data in the database to a
        load case file.  The database remains unchanged by this operation.  A
        pointer is also established to the written set of results on the load
        case file.  This pointer (LCNO) can then be used on the LCASE or LCOPER
        commands to read the load case data into the database.  By default,
        only summable results data (such as displacements, stresses, elastic
        strains) and constant results data (such as volume) are written to the
        load case file unless requested (LCSUM command).  Non-summable results
        data (such as plastic strains, strain energy), boundary conditions, and
        nodal loads are not written to the load case file.  The load case file
        may be named by default or by a user name.  Rewriting to the same file
        overwrites the previous data.  See the LCDEF command for status and
        erase operations.
        """
        command = f"LCWRITE,{lcno},{fname},{ext}"
        return self.run(command, **kwargs)

    def lczero(self, **kwargs):
        """Zeroes the results portion of the database.

        APDL Command: LCZERO

        Notes
        -----
        Often used before the LCOPER command.  Same as LCOPER,ZERO.
        """
        command = f"LCZERO,"
        return self.run(command, **kwargs)

    def rappnd(self, lstep="", time="", **kwargs):
        """Appends results data from the database to the results file.

        APDL Command: RAPPND

        Parameters
        ----------
        lstep
            Load step number to be assigned to the results data set.  If it is
            the same as an existing load step number on the results file, the
            appended load step will be inaccessible.  Defaults to 1.

        time
            Time value to be assigned to the results data set.  Defaults to
            0.0.  A time value greater than the last load step should be used.

        Notes
        -----
        This command is typically used to append the results from a load case
        combination to the results file.  See the LCWRITE command to create a
        separate load case file.  Only summable and constant data are written
        to the results file by default; non-summable data are not written
        unless requested (LCSUM command). RAPPND should not be used to append
        results from a harmonic analysis.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"RAPPND,{lstep},{time}"
        return self.run(command, **kwargs)



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Miscellaneous methods not covered in the documentation."""


class Misc:
    def verify(self, case="", level="", **kwargs):
        """Enter the verification run mode.

        .. note::
           This command is only valid at the ``/BEGIN`` level, obtained
           with ``mapdl.finish()``.

        Parameters
        ----------
        case : str, optional
            Optional title of the verification manual file.  Also accepts
            ``'OFF'`` to disable the verification run mode.

        level : int, optional
            Verification level ranging from 1 to 6 defaulting to 4.

        Returns
        -------
        str
            Command output.

        Examples
        --------
        Enter the verification routine with the default option.

        >>> mapdl.finish()
        >>> mapdl.verify('VM1')
        '*** VERIFICATION RUN - CASE VM1                              ***  OPTION=  4'

        """
        return self.run(f"/VERIFY,{case},{level}", **kwargs)



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from functools import wraps
import re

import numpy as np

from ._commands import (
    apdl,
    aux2_,
    aux3_,
    aux12_,
    aux15_,
    conn,
    database,
    display_,
    graphics_,
    hidden,
    inq_func,
    map_cmd,
    misc,
    post1_,
    post26_,
    preproc,
    reduced,
    session,
    solution,
)

# compiled regular expressions used for parsing tablular outputs
REG_LETTERS = re.compile(r"[a-df-zA-DF-Z]+")  # all except E or e
REG_FLOAT_INT = re.compile(
    r"[+-]?[0-9]*[.]?[0-9]*[Ee]?[+-]?[0-9]+|\s[0-9]+\s"
)  # match number groups
BC_REGREP = re.compile(r"^\s*([0-9]+)\s*([A-Za-z]+)((?:\s+[0-9]*[.]?[0-9]+)+)$")


MSG_NOT_PANDAS = """'Pandas' is not installed or could not be found.
Hence this command is not applicable.

You can install it using:
pip install pandas
"""

MSG_BCLISTINGOUTPUT_TO_ARRAY = """This command has strings values in some of its columns (such 'UX', 'FX', 'UY', 'TEMP', etc),
so it cannot be converted to Numpy Array.

Please use 'to_list' or 'to_dataframe' instead."""


# Identify where the data start in the output
GROUP_DATA_START = ["NODE", "ELEM"]

# Allowed commands to get output as array or dataframe.
# In theory, these commands should follow the same format.
# Some of them are not documented (already deprecated?)
# So they are not in the Mapdl class,
# so they won't be wrapped.
CMD_RESULT_LISTING = [
    "NLIN",  # not documented
    "PRCI",
    "PRDI",  # Not documented.
    "PREF",  # Not documented.
    "PREN",
    "PRER",
    "PRES",
    "PRET",
    "PRGS",  # Not documented.
    "PRIN",
    "PRIT",
    "PRJS",
    "PRNL",
    "PRNM",  # Not documented.
    "PRNS",
    "PROR",
    "PRPA",
    "PRRF",
    "PRRS",
    "PRSE",
    "PRSS",  # Not documented.
    "PRST",  # Not documented.
    "PRVE",
    "PRXF",  # Not documented.
    "SWLI",
]

CMD_BC_LISTING = [
    "DKLI",
    "DLLI",
    "DALI",
    "DLIS",
    "FKLI",
    "FLIS",
    "SFLL",
    # "SFAL",   Define two integers before label (regex)
    # "SFLI",   Use two lines to define each BC in the list
    # "SFEL",   Use two lines to define each BC in the list
    "BFKL",
    "BFLL",
    "BFAL",
]

COLNAMES_BC_LISTING = {
    "DKLI": ["KEYPOINT", "LABEL", "REAL", "IMAG", "EXP KEY"],
    "DLLI": ["LINE", "LABEL", "REAL", "IMAG", "NAREA"],
    "DALI": ["AREA", "LABEL", "REAL", "IMAG"],
    "DLIS": ["NODE", "LABEL", "REAL", "IMAG"],
    "FKLI": ["KEYPOINT", "LABEL", "REAL", "IMAG"],
    "FLIS": ["NODE", "LABEL", "REAL", "IMAG"],
    "SFLL": ["LINE", "LABEL", "VALI", "VALJ", "VAL2I", "VAL2J"],
    "BFKL": ["KEYPOINT", "LABEL", "VALUE"],
    "BFLL": ["LINE", "LABEL", "VALUE"],
    "BFAL": ["AREA", "LABEL", "VALUE"],
}

CMD_ENTITY_LISTING = [
    "NLIS",
    # "ELIS", # To be implemented later
    # "KLIS",
    # "LLIS",
    # "ALIS",
    # "VLIS",
]

CMD_LISTING = []
CMD_LISTING.extend(CMD_ENTITY_LISTING)
CMD_LISTING.extend(CMD_RESULT_LISTING)

# Adding empty lines to match current format.
CMD_DOCSTRING_INJECTION = r"""
Returns
-------

str
    Str object with the command console output.

    This object also has the extra methods:
    :meth:`to_list() <ansys.mapdl.core.commands.CommandListingOutput.to_list>`,
    :meth:`to_array() <ansys.mapdl.core.commands.CommandListingOutput.to_array>` (only on listing commands) and
    :meth:`to_dataframe() <ansys.mapdl.core.commands.CommandListingOutput.to_dataframe>` (only if Pandas is installed).
    |bl|
    **NOTE**: If you use these methods, you might
    obtain a lower precision than using :class:`Mesh <ansys.mapdl.core.mesh_grpc.MeshGrpc>` methods.
    |bl|
    For more information visit :ref:`user_guide_postprocessing`.

"""

XSEL_DOCSTRING_INJECTION = r"""
Returns
-------

np.ndarray
    Numpy array with the ids of the selected entities.

    For more information visit :ref:`user_guide_postprocessing`.

"""


CMD_XSEL = [
    "NSEL",
    "ESEL",
    "KSEL",
    "LSEL",
    "ASEL",
    "VSEL",
]


def get_indentation(indentation_regx, docstring):
    return re.findall(indentation_regx, docstring, flags=re.DOTALL | re.IGNORECASE)[0][
        0
    ]


def indent_text(indentation, docstring_injection):
    return "\n".join(
        [
            indentation + each
            for each in docstring_injection.splitlines()
            if each.strip()
        ]
    )


def get_docstring_indentation(docstring):
    indentation_regx = r"\n(\s*)\n"
    return get_indentation(indentation_regx, docstring)


def get_sections(docstring):
    return [
        each.strip().lower() for each in re.findall(r"\n\s*(\S*)\n\s*-+\n", docstring)
    ]


def get_section_indentation(section_name, docstring):
    sections = get_sections(docstring)
    if section_name.lower().strip() not in sections:
        raise ValueError(
            f"This section '{section_name.lower().strip()}' does not exist in the docstring."
        )
    section_match = section_name + r"\n\s*-*"
    indent_match = r"\n(\s*)(\S)"
    indentation_regx = section_match + indent_match
    return get_indentation(indentation_regx, docstring)


def inject_before(section, indentation, indented_doc_inject, docstring):
    return re.sub(
        section + r"\n\s*-*",
        f"{indented_doc_inject.strip()}\n\n{indentation}" + r"\g<0>",
        docstring,
        flags=re.IGNORECASE,
    )


def inject_after_return_section(indented_doc_inject, docstring):
    return re.sub(
        "Returns" + r"\n\s*-*",
        f"{indented_doc_inject.strip()}\n",
        docstring,
        flags=re.IGNORECASE,
    )


def inject_docs(docstring, docstring_injection=None):
    """Inject a string in a docstring"""
    if not docstring_injection:
        docstring_injection = CMD_DOCSTRING_INJECTION

    return_header = r"Returns\n\s*-*"

    if docstring_injection.splitlines()[-2].strip() in docstring:
        # In case the docstring already has the injection.
        return docstring
    elif re.search(return_header, docstring):
        # There is a return block already, probably it should not.
        indentation = get_section_indentation("Returns", docstring)
        indented_doc_inject = indent_text(indentation, docstring_injection)
        return inject_after_return_section(indented_doc_inject, docstring)
    else:
        # There is not returns header
        # find sections
        sections = get_sections(docstring)

        if "parameters" in sections:
            ind = sections.index("parameters")
            if ind == len(sections) - 1:
                # The parameters is the last bit. Just append it.
                indentation = get_section_indentation("Parameters", docstring)
                indented_doc_inject = indent_text(indentation, docstring_injection)
                return docstring + "\n" + indented_doc_inject
            else:
                # inject it right before the section after 'parameter'
                sect_after_parameter = sections[ind + 1]
                indentation = get_section_indentation(sect_after_parameter, docstring)
                indented_doc_inject = indent_text(indentation, docstring_injection)
                return inject_before(
                    sect_after_parameter,
                    indentation,
                    indented_doc_inject,
                    docstring,
                )

        elif "notes" in sections:
            indentation = get_section_indentation("Notes", docstring)
            indented_doc_inject = indent_text(indentation, docstring_injection)
            return inject_before("Notes", indentation, indented_doc_inject, docstring)

        else:
            indentation = get_docstring_indentation(docstring)
            indented_doc_inject = indent_text(indentation, docstring_injection)
            return docstring + "\n" + indented_doc_inject


def check_valid_output(func):
    """Wrapper that check if output can be wrapped by pandas, if not, it will raise an exception."""

    @wraps(func)
    def func_wrapper(self, *args, **kwargs):
        output = self.__str__()
        if (
            "*** WARNING ***" in output or "*** ERROR ***" in output
        ):  # Error should be caught in mapdl.run.
            err_type = re.findall(r"*** (.*) ***", output)[0]
            msg = f"Unable to parse because of next {err_type.title()}" + "\n".join(
                output.splitlines()[-2:]
            )
            raise ValueError(msg)
        else:
            return func(self, *args, **kwargs)

    return func_wrapper


class PreprocessorCommands(
    preproc.database.Database,
    preproc.explicit_dynamics.ExplicitDynamics,
    preproc.lines.Lines,
    preproc.areas.Areas,
    preproc.nodes.Nodes,
    preproc.keypoints.KeyPoints,
    preproc.artificially_matched_layers.ArtificiallyMatchedLayers,
    preproc.booleans.Booleans,
    preproc.constraint_equations.ConstraintEquations,
    preproc.coupled_dof.CoupledDOF,
    preproc.real_constants.RealConstants,
    preproc.digitizing.Digitizing,
    preproc.element_type.ElementType,
    preproc.elements.Elements,
    preproc.hard_points.HardPoints,
    preproc.material_data_tables.MaterialDataTables,
    preproc.meshing.Meshing,
    preproc.morphing.Morphing,
    preproc.materials.Materials,
    preproc.primitives.Primitives,
    preproc.sections.Sections,
    preproc.special_purpose.SpecialPurpose,
    preproc.status.Status,
    preproc.superelements.Superelements,
    preproc.volumes.Volumes,
):
    pass


class APDLCommands(
    apdl.abbreviations.Abbreviations,
    apdl.array_param.ArrayParam,
    apdl.macro_files.MacroFiles,
    apdl.matrix_op.MatrixOP,
    apdl.parameter_definition.ParameterDefinition,
    apdl.process_controls.ProcessControls,
):
    pass


class Aux2Commands(aux2_.bin_dump.BinDump, aux2_.bin_manip.BinManip):
    pass


class Aux12Commands(
    aux12_.radiosity_solver.RadiositySolver,
    aux12_.radiation_mat.RadiationMat,
    aux12_.general_radiation.GeneralRadiation,
):
    pass


class DatabaseCommands(
    database.setup.Setup,
    database.picking.Picking,
    database.coord_sys.CoordinateSystem,
    database.selecting.Selecting,
    database.working_plane.WorkingPlane,
    database.components.Components,
):
    pass


class DisplayCommands(display_.setup.Setup):
    pass


class GraphicsCommands(
    graphics_.annotation.Annotation,
    graphics_.graphs.Graphs,
    graphics_.labeling.Labelling,
    graphics_.scaling.Scaling,
    graphics_.setup.Setup,
    graphics_.style.Style,
    graphics_.views.Views,
):
    pass


class MiscCommands(misc.misc.Misc):
    pass


class Post1Commands(
    post1_.animation.Animation,
    post1_.controls.Controls,
    post1_.element_table.ElementTable,
    post1_.failure_criteria.FailureCriteria,
    post1_.listing.Listing,
    post1_.load_case.LoadCase,
    post1_.magnetics_calc.MagneticsCalc,
    post1_.path_operations.PathOperations,
    post1_.results.Results,
    post1_.setup.Setup,
    post1_.special.Special,
    post1_.status.Status,
    post1_.surface_operations.SurfaceOperations,
    post1_.trace_points.TracePoints,
):
    pass


class Post26Commands(
    post26_.controls.Controls,
    post26_.display.Display,
    post26_.listing.Listing,
    post26_.operations.Operations,
    post26_.setup.Setup,
    post26_.special.Special,
    post26_.status.Status,
):
    pass


class ReducedCommands(
    reduced.generation.Generation,
    reduced.preparation.Preparation,
    reduced.setup.Setup,
    reduced.use_pass.UsePass,
):
    pass


class SessionCommands(
    session.files.Files,
    session.list_controls.ListControls,
    session.processor_entry.ProcessorEntry,
    session.run_controls.RunControls,
):
    pass


class SolutionCommands(
    solution.analysis_options.AnalysisOptions,
    solution.birth_and_death.BirthAndDeath,
    solution.dynamic_options.DynamicOptions,
    solution.fe_body_loads.FeBodyLoads,
    solution.fe_constraints.FeConstraints,
    solution.fe_forces.FeForces,
    solution.fe_surface_loads.FeSurfaceLoads,
    solution.gap_conditions.GapConditions,
    solution.inertia.Inertia,
    solution.load_step_operations.LoadStepOperations,
    solution.load_step_options.LoadStepOptions,
    solution.master_dof.MasterDOF,
    solution.miscellaneous_loads.MiscellaneousLoads,
    solution.multi_field_solver_convergence_controls.MultiFieldConvergenceControls,
    solution.multi_field_solver_definition_commands.MultiFieldSolverDefinitionCommands,
    solution.multi_field_solver_global_controls.MultiFieldSolverGlobalControls,
    solution.multi_field_solver_interface_mapping.MultiFieldSolverInterfaceMapping,
    solution.multi_field_solver_load_transfer.MultiFieldSolverLoadTransfer,
    solution.multi_field_solver_time_controls.MultiFieldSolverTimeControls,
    solution.nonlinear_options.NonLinearOptions,
    solution.ocean.Ocean,
    solution.radiosity.Radiosity,
    solution.rezoning.Rezoning,
    solution.solid_body_loads.SolidBodyLoads,
    solution.solid_constraints.SolidConstraints,
    solution.solid_forces.SolidForces,
    solution.solid_surface_loads.SolidSurfaceLoads,
    solution.solution_status.SolutionStatus,
    solution.spectrum_options.SpectrumOptions,
    solution.twod_to_3d_analysis.TwoDTo3DAnalysis,
):
    pass


class InqFunctions(inq_func.inq_function):
    pass


class Commands(
    APDLCommands,
    Aux2Commands,
    Aux12Commands,
    DatabaseCommands,
    DisplayCommands,
    GraphicsCommands,
    MiscCommands,
    Post1Commands,
    Post26Commands,
    PreprocessorCommands,
    ReducedCommands,
    SessionCommands,
    SolutionCommands,
    aux3_.Aux3,
    aux15_.Aux15,
    conn.Conn,
    hidden._Hidden,
    map_cmd.MapCommand,
    InqFunctions,
):
    """Wrapped MAPDL commands"""


def _requires_pandas(func):
    """Wrapper that check ``HAS_PANDAS``, if not, it will raise an exception."""

    def func_wrapper(self, *args, **kwargs):
        if HAS_PANDAS:
            return func(self, *args, **kwargs)
        else:
            raise ModuleNotFoundError(MSG_NOT_PANDAS)

    return func_wrapper


class CommandOutput(str):
    """Custom string subclass for handling the commands output.

    This class is a subclass of python :class`str`, hence it has all the methods of
    a string python object.

    Additionally it provides the following attributes:

    * :attr:`cmd() <ansys.mapdl.core.commands.CommandOutput.cmd>`
    * :attr:`command() <ansys.mapdl.core.commands.CommandOutput.command>`

    """

    ## References:
    # - https://stackoverflow.com/questions/7255655/how-to-subclass-str-in-python
    # - https://docs.python.org/3/library/collections.html#userstring-objects
    # - Source code of UserString

    def __new__(cls, content, cmd=None):
        obj = super().__new__(cls, content)
        obj._cmd = cmd
        return obj

    @property
    def cmd(self):
        """Cached original command to generate this command output."""
        return self._cmd.split(",")[0]

    @cmd.setter
    def cmd(self, cmd):
        """Not allowed to change the value of ``cmd``."""
        raise AttributeError("The `cmd` attribute cannot be set")

    @property
    def command(self):
        return self._cmd

    @command.setter
    def command(self):
        """Not allowed to change the value of ``command``."""
        pass


class CommandListingOutput(CommandOutput):
    """Allow the conversion of command output to native Python types.

    Custom class for handling the commands whose output is sensible to be converted to
    a list of lists, a Numpy array or a Pandas DataFrame.

    This class is a subclass of python :class:`str`, hence it has all the methods of
    a string python object.

    Additionally it provides the following methods:

    * :func:`to_list() <ansys.mapdl.core.commands.CommandListingOutput.to_list>`
    * :func:`to_array() <ansys.mapdl.core.commands.CommandListingOutput.to_array>`
    * :func:`to_dataframe() <ansys.mapdl.core.commands.CommandListingOutput.to_dataframe>`

    """

    def __new__(cls, content, cmd=None, magicwords=None, columns_names=None):
        obj = super().__new__(cls, content)
        obj._cmd = cmd
        obj._magicwords = magicwords
        obj._columns_names = columns_names
        return obj

    def __init__(self, *args, **kwargs):
        self._cache = None

    def _is_data_start(self, line, magicwords=None):
        """Check if line is the start of a data group."""
        if not magicwords:
            if self._magicwords:
                magicwords = self._magicwords
            else:
                magicwords = GROUP_DATA_START

        # Checking if we are supplying a custom start function.
        if self.custom_data_start(line) is not None:
            return self.custom_data_start(line)

        if line.split():
            if line.split()[0] in magicwords or self.custom_data_start(line):
                return True
        return False

    def _is_data_end(self, line):
        """Check if line is the end of a data group."""

        # Checking if we are supplying a custom start function.
        if self.custom_data_end(line) is not None:
            return self.custom_data_end(line)
        else:
            return self._is_empty(line)

    def custom_data_start(self, line):
        """Custom data start line check function.

        This function is left empty so it can be overwritten by the user.

        If modified, it should return ``True`` when the line is the start
        of a data group, otherwise it should return ``False``.
        """
        return None

    def custom_data_end(self, line):
        """Custom data end line check function.

        This function is left empty so it can be overwritten by the user.

        If modified, it should return ``True`` when the line is the end
        of a data group, otherwise it should return ``False``.
        """
        return None

    @staticmethod
    def _is_empty_line(line):
        return bool(line.split())

    def _format(self):
        """Perform some formatting (replacing mainly) in the raw text."""
        return re.sub(r"[^E](-)", " -", self.__str__())

    def _get_body(self, trail_header=None):
        """Get command body text.

        It removes the maximum absolute values tail part and makes sure there is
        separation between columns.
        """
        # Doing some formatting of the string
        body = self._format().splitlines()

        if not trail_header:
            trail_header = ["MAXIMUM ABSOLUTE VALUES", "TOTAL VALUES"]

        if not isinstance(trail_header, list):
            trail_header = list(trail_header)

        # Removing parts matching trail_header
        for each_trail_header in trail_header:
            if each_trail_header in self.__str__():
                # starting to check from the bottom.
                for i in range(len(body) - 1, -1, -1):
                    if each_trail_header in body[i]:
                        break
                body = body[:i]
        return body

    def _get_data_group_indexes(self, body, magicwords=None):
        """Return the indexes of the start and end of the data groups."""
        if "*****ANSYS VERIFICATION RUN ONLY*****" in str(self[:1000]):
            shift = 2
        else:
            shift = 0

        # Getting pairs of starting end
        start_idxs = [
            ind
            for ind, each in enumerate(body)
            if self._is_data_start(each, magicwords=magicwords)
        ]
        end_idxs = [
            ind - shift for ind, each in enumerate(body) if self._is_empty_line(each)
        ]

        indexes = [*start_idxs, *end_idxs]
        indexes.sort()

        ends = [indexes[indexes.index(each) + 1] for each in start_idxs[:-1]]
        ends.append(len(body))

        return zip(start_idxs, ends)

    def get_columns(self):
        """Get the column names for the dataframe.

        Returns
        -------
        List of strings

        """
        if self._columns_names:
            return self._columns_names

        body = self._get_body()
        pairs = list(self._get_data_group_indexes(body))
        try:
            return body[pairs[0][0]].split()
        except:
            return None

    def _parse_table(self):
        """Parse tabular command output.

        Returns
        -------
        numpy.ndarray
            Parsed tabular data from command output.

        """
        parsed_lines = []
        for line in self.splitlines():
            # exclude any line containing characters [A-Z] except for E
            if line.strip() and not REG_LETTERS.search(line):
                items = REG_FLOAT_INT.findall(line)
                if items:
                    parsed_lines.append(items)
        return np.array(parsed_lines, dtype=np.float64)

    @property
    def _parsed(self):
        """Return parsed output."""
        if self._cache is None:
            self._cache = self._parse_table()
        return self._cache

    @check_valid_output
    def to_list(self):
        """Export the command output a list or list of lists.

        Returns
        -------
        list
        """
        return self._parsed.tolist()

    def to_array(self):
        """Export the command output as a numpy array.

        Returns
        -------
        numpy.ndarray
            Numpy array of floats.
        """
        return self._parsed

    def to_dataframe(self, data=None, columns=None):
        """Export the command output as a Pandas DataFrame.

        Parameters
        ----------
        data : numpy.ndarray (structured or homogeneous), Iterable, dict, or DataFrame
            The data to be converted to the dataframe values.  Passed directly
            to the pandas.DataFrame constructor.  Dict can contain Series,
            arrays, constants, dataclass or list-like objects. If data is a
            dict, column order follows insertion-order.

        columns : Index or array-like
            Iterable with columns names.  Passed directly to the
            pandas.DataFrame constructor.  Column labels to use for resulting
            frame when data does not have them, defaulting to RangeIndex(0, 1,
            2, ..., n). If data contains column labels, will perform column
            selection instead.

        Returns
        -------
        pandas.DataFrame
            Pandas DataFrame

        Notes
        -----
        The returned dataframe has all its data converted to float
        (inheritate from :func:`to_array()
        <ansys.mapdl.core.commands.CommandListingOutput.to_array>` method).
        """
        try:
            import pandas as pd
        except ModuleNotFoundError:
            raise ModuleNotFoundError(MSG_NOT_PANDAS)

        if data is None:
            data = self.to_list()
        if not columns:
            columns = self.get_columns()

        return pd.DataFrame(data=data, columns=columns)


class BoundaryConditionsListingOutput(CommandListingOutput):
    """Allow the conversion of command output to native Python types.

    Custom class for handling the boundary condition listing commands
    whose output is sensible to be converted to a list of lists,
    or a Pandas DataFrame.

    This class is a subclass of python :class:`str`, hence it has all the methods of
    a string python object and it can be used as such.

    Additionally it provides the following methods:

    * :func:`to_list() <ansys.mapdl.core.commands.BoundaryConditionsListingOutput.to_list>`
    * :func:`to_dataframe() <ansys.mapdl.core.commands.BoundaryConditionsListingOutput.to_dataframe>`

    """

    def bc_colnames(self):
        """Get the column names based on bc list command"""

        bc_type = {
            "BODY FORCES": "BF",
            "SURFACE LOAD": "SF",
            "POINT LOAD": "F",
            "FORCES": "F",
            "CONSTRAINTS": "D",
        }

        entity = {
            "KEYPOINT": "K",
            "LINE": "L",
            "AREA": "A",
            "NODE": "",
            "ELEMENT": "E",
        }

        title = self._get_body()[0]

        _bcType = [i for i in bc_type.keys() if i in title]
        _entity = [i for i in entity.keys() if i in title]

        if _bcType and _entity:

            key_bc = bc_type[_bcType[0]] + entity[_entity[0]] + "LIST"
            key_bc = key_bc[:4]

            if key_bc in COLNAMES_BC_LISTING.keys():

                _cols = COLNAMES_BC_LISTING[key_bc]

                # Check num columns in data
                ldata = []
                for line in self.splitlines():
                    line = line.strip()
                    # exclude any line containing characters [A-Z] except for E
                    if line:
                        items = BC_REGREP.findall(line)
                        if items:
                            ldata = list(items[0][:2]) + items[0][2].split()
                            break

                if ldata:
                    if len(_cols) > len(ldata):
                        _cols = _cols[: len(ldata)]

                    return _cols

        return None

    def get_columns(self):
        """Get the column names for the dataframe.

        Returns
        -------
        List of strings

        """
        if self._columns_names:
            return self._columns_names

        bc_colnames = self.bc_colnames()

        if bc_colnames:
            return bc_colnames

        body = self._get_body()

        pairs = list(self._get_data_group_indexes(body))
        try:
            return body[pairs[0][0]].split()
        except:
            return None

    def _parse_table(self):
        """Parse tabular command output."""
        parsed_lines = []
        for line in self.splitlines():
            line = line.strip()
            # exclude any line containing characters [A-Z] except for E
            if line:
                items = BC_REGREP.findall(line)
                if items:
                    parsed_lines.append(list(items[0][:2]) + items[0][2].split())

        return parsed_lines

    @check_valid_output
    def to_list(self):
        """Export the command output a list or list of lists.

        Returns
        -------
        list
        """
        return self._parsed

    def to_array(self):
        raise ValueError(MSG_BCLISTINGOUTPUT_TO_ARRAY)

    def to_dataframe(self):
        """Convert the command output to a Pandas Dataframe.

        Returns
        -------
        pandas.DataFrame
            Pandas Dataframe

        Notes
        -----

        If present, the next columns will be converted to:

        * ``'NODE'``: int
        * ``'LABEL'``: str
        * ``'REAL'``: float
        * ``'IMAG'``: float

        """
        df = super().to_dataframe(data=self.to_list())

        primitives = ["KEYPOINT", "LINE", "AREA", "VOLUME", "NODE", "ELEMENT"]

        float_col = ["REAL", "IMAG", "VALUE", "VALI", "VALJ"]

        for i in df.columns.intersection(primitives):
            df[i] = df[i].astype(np.int32, copy=False)

        if "LABEL" in df.columns:
            df["LABEL"] = df["LABEL"].astype(str, copy=False)

        for i in df.columns.intersection(float_col):
            df[i] = df[i].astype(np.float64, copy=False)

        return df


class ComponentListing(CommandListingOutput):
    @property
    def _parsed(self):
        from ansys.mapdl.core.component import _parse_cmlist

        # To keep same API as commands
        return np.array(list(_parse_cmlist(self).keys()))


class StringWithLiteralRepr(str):
    def __repr__(self):
        return self.__str__()



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""These SESSION commands are used to enter and exit the various
processors in the program.
"""


class ProcessorEntry:
    def aux2(self, **kwargs):
        """Enters the binary file dumping processor.

        APDL Command: /AUX2

        Notes
        -----
        Enters the binary file dumping processor (ANSYS auxiliary processor
        AUX2).  This processor is used to dump the contents of certain ANSYS
        binary files for visual examination.

        This command is valid only at the Begin Level.
        """
        command = "/AUX2,"
        return self.run(command, **kwargs)

    def aux3(self, **kwargs):
        """Enters the results file editing processor.

        APDL Command: /AUX3

        Notes
        -----
        Enters the results file editing processor (ANSYS auxiliary processor
        AUX3).  This processor is used to edit ANSYS results files.

        This command is valid only at the Begin Level.
        """
        command = "/AUX3,"
        return self.run(command, **kwargs)

    def aux12(self, **kwargs):
        """Enters the radiation processor.

        APDL Command: /AUX12

        Notes
        -----
        Enters the radiation processor (ANSYS auxiliary processor AUX12).  This
        processor supports the Radiation Matrix and the Radiosity Solver
        methods.

        This command is valid only at the Begin Level.
        """
        command = "/AUX12,"
        return self.run(command, **kwargs)

    def aux15(self, **kwargs):
        """Enters the IGES file transfer processor.

        APDL Command: /AUX15

        Notes
        -----
        Enters the IGES file transfer processor (ANSYS auxiliary processor
        AUX15), used to read an IGES data file into the ANSYS program.

        This command is valid only at the Begin Level.
        """
        command = "/AUX15,"
        return self.run(command, **kwargs)

    def finish(self, **kwargs):
        """Exits normally from a processor.

        APDL Command: FINISH

        Notes
        -----
        Exits any of the ANSYS processors or the DISPLAY program.  For the
        ANSYS processors, data will remain intact in the database but the
        database is not automatically written to a file (use the SAVE command
        to write the database to a file).  See also the /QUIT command for an
        alternate processor exit command.  If exiting POST1, POST26, or OPT,
        see additional notes below.

        POST1:  Data in the database will remain intact, including the POST1
        element table data, the path table data, the fatigue table data, and
        the load case pointers.

        POST26:  Data in the database will remain intact, except that POST26
        variables are erased and specification commands (such as FILE, PRTIME,
        NPRINT, etc.) are reset.  Use the /QUIT command to exit the processor
        and bypass these exceptions.

        This command is valid in any processor.  This command is not valid at
        the Begin level.
        """
        command = "FINISH,"
        return self.run(command, **kwargs)

    def post1(self, **kwargs):
        """Enters the database results postprocessor.

        APDL Command: /POST1

        Notes
        -----
        Enters the general database results postprocessor (POST1).  All load
        symbols (/PBC, /PSF, or /PBF) are automatically turned off with this
        command.

        This command is valid only at the Begin Level.
        """
        command = "/POST1,"
        return self.run(command, **kwargs)

    def post26(self, **kwargs):
        """Enters the time-history results postprocessor.

        APDL Command: /POST26

        Notes
        -----
        Enters the time-history results postprocessor (POST26).

        This command is valid only at the Begin Level.
        """
        command = "/POST26,"
        return self.run(command, **kwargs)

    def prep7(self, **kwargs):
        """Enters the model creation preprocessor.

        APDL Command: /PREP7

        Notes
        -----
        Enters the general input data preprocessor (PREP7).

        This command is valid only at the Begin Level.
        """
        command = "/PREP7,"
        return self.run(command, **kwargs)

    def quit(self, **kwargs):
        """Exits a processor.

        APDL Command: /QUIT

        Notes
        -----
        This is an alternative to the FINISH command.  If any cleanup or file
        writing is normally done by the FINISH command, it is bypassed if the
        /QUIT command is used instead.  A new processor may be entered after
        this command.  See the /EXIT command to terminate the run.

        This command is valid in any processor.  This command is not valid at
        the Begin level.
        """
        command = "/QUIT,"
        return self.run(command, **kwargs)

    def slashsolu(self, **kwargs):
        """Enters the solution processor.

        APDL Command: /SOLU

        Notes
        -----
        This command is valid only at the Begin Level.
        """
        command = "/SOLU,"
        return self.run(command, **kwargs)
